import re
from numpy.random import shuffle
from copy import deepcopy
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.rdchem import ChiralType

VERBOSE = False
USE_STEREOCHEMISTRY = True
MAXIMUM_NUMBER_UNMAPPED_PRODUCT_ATOMS = 5
INCLUDE_ALL_UNMAPPED_REACTANT_ATOMS = True

def mols_from_smiles_list(all_smiles):
    '''Given a list of smiles strings, this function creates rdkit
    molecules'''
    mols = []
    for smiles in all_smiles:
        if not smiles: continue
        mols.append(Chem.MolFromSmiles(smiles))
    return mols

def replace_deuterated(smi):
    return re.sub('\[2H\]', r'[H]', smi)

def clear_mapnum(mol):
    [a.ClearProp('molAtomMapNumber') for a in mol.GetAtoms() if a.HasProp('molAtomMapNumber')]
    return mol

def get_tagged_atoms_from_mols(mols):
    '''Takes a list of RDKit molecules and returns total list of
    atoms and their tags'''
    atoms = []
    atom_tags = []
    for mol in mols:
        new_atoms, new_atom_tags = get_tagged_atoms_from_mol(mol)
        atoms += new_atoms 
        atom_tags += new_atom_tags
    return atoms, atom_tags

def get_tagged_atoms_from_mol(mol):
    '''Takes an RDKit molecule and returns list of tagged atoms and their
    corresponding numbers'''
    atoms = []
    atom_tags = []
    for atom in mol.GetAtoms():
        if atom.HasProp('molAtomMapNumber'):
            atoms.append(atom)
            atom_tags.append(str(atom.GetProp('molAtomMapNumber')))
    return atoms, atom_tags

def atoms_are_different(atom1, atom2):
    '''Compares two RDKit atoms based on basic properties'''

    if atom1.GetSmarts() != atom2.GetSmarts(): return True # easiest check
    if atom1.GetAtomicNum() != atom2.GetAtomicNum(): return True # must be true for atom mapping
    if atom1.GetTotalNumHs() != atom2.GetTotalNumHs(): return True
    if atom1.GetFormalCharge() != atom2.GetFormalCharge(): return True
    if atom1.GetDegree() != atom2.GetDegree(): return True
    #if atom1.IsInRing() != atom2.IsInRing(): return True # do not want to check this!
    # e.g., in macrocycle formation, don't want the template to include the entire ring structure
    if atom1.GetNumRadicalElectrons() != atom2.GetNumRadicalElectrons(): return True
    if atom1.GetIsAromatic() != atom2.GetIsAromatic(): return True 

    # Check bonds and nearest neighbor identity
    bonds1 = sorted([bond_to_label(bond) for bond in atom1.GetBonds()]) 
    bonds2 = sorted([bond_to_label(bond) for bond in atom2.GetBonds()]) 
    if bonds1 != bonds2: return True

    return False

def find_map_num(mol, mapnum):
    return [(a.GetIdx(), a) for a in mol.GetAtoms() if a.HasProp('molAtomMapNumber') 
         and a.GetProp('molAtomMapNumber') == str(mapnum)][0]

def get_tetrahedral_atoms(reactants, products):
    tetrahedral_atoms = []
    for reactant in reactants:
        for ar in reactant.GetAtoms():
            if not ar.HasProp('molAtomMapNumber'):
                continue
            atom_tag = ar.GetProp('molAtomMapNumber')
            ir = ar.GetIdx()
            for product in products:
                try:
                    (ip, ap) = find_map_num(product, atom_tag)
                    if ar.GetChiralTag() != ChiralType.CHI_UNSPECIFIED or\
                            ap.GetChiralTag() != ChiralType.CHI_UNSPECIFIED:
                        tetrahedral_atoms.append((atom_tag, ar, ap))
                except IndexError:
                    pass
    return tetrahedral_atoms

def set_isotope_to_equal_mapnum(mol):
    for a in mol.GetAtoms():
        if a.HasProp('molAtomMapNumber'):
            a.SetIsotope(int(a.GetProp('molAtomMapNumber')))
            
def get_frag_around_tetrahedral_center(mol, idx):
    '''Builds a MolFragment using neighbors of a tetrahedral atom,
    where the molecule has already been updated to include isotopes'''
    ids_to_include = [idx]
    for neighbor in mol.GetAtomWithIdx(idx).GetNeighbors():
        ids_to_include.append(neighbor.GetIdx())
    symbols = ['[{}{}]'.format(a.GetIsotope(), a.GetSymbol()) if a.GetIsotope() != 0\
               else '[#{}]'.format(a.GetAtomicNum()) for a in mol.GetAtoms()]
    return Chem.MolFragmentToSmiles(mol, ids_to_include, isomericSmiles=True,
                                   atomSymbols=symbols, allBondsExplicit=True,
                                   allHsExplicit=True)
            
def check_tetrahedral_centers_equivalent(atom1, atom2):
    '''Checks to see if tetrahedral centers are equivalent in
    chirality, ignoring the ChiralTag. Owning molecules of the
    input atoms must have been Isotope-mapped'''
    atom1_frag = get_frag_around_tetrahedral_center(atom1.GetOwningMol(), atom1.GetIdx())
    atom1_neighborhood = Chem.MolFromSmiles(atom1_frag, sanitize=False)
    for matched_ids in atom2.GetOwningMol().GetSubstructMatches(atom1_neighborhood, useChirality=True):
        if atom2.GetIdx() in matched_ids:
            return True
    return False

def clear_isotope(mol):
    [a.SetIsotope(0) for a in mol.GetAtoms()]

def get_changed_atoms(reactants, products):
    '''Looks at mapped atoms in a reaction and determines which ones changed'''

    err = 0
    prod_atoms, prod_atom_tags = get_tagged_atoms_from_mols(products)

    if VERBOSE: print('Products contain {} tagged atoms'.format(len(prod_atoms)))
    if VERBOSE: print('Products contain {} unique atom numbers'.format(len(set(prod_atom_tags))))

    reac_atoms, reac_atom_tags = get_tagged_atoms_from_mols(reactants)
    if len(set(prod_atom_tags)) != len(set(reac_atom_tags)):
        if VERBOSE: print('warning: different atom tags appear in reactants and products')
        #err = 1 # okay for Reaxys, since Reaxys creates mass
    if len(prod_atoms) != len(reac_atoms):
        if VERBOSE: print('warning: total number of tagged atoms differ, stoichometry != 1?')
        #err = 1

    # Find differences 
    changed_atoms = [] # actual reactant atom species
    changed_atom_tags = [] # atom map numbers of those atoms

    # Product atoms that are different from reactant atom equivalent
    for i, prod_tag in enumerate(prod_atom_tags):

        for j, reac_tag in enumerate(reac_atom_tags):
            if reac_tag != prod_tag: continue
            if reac_tag not in changed_atom_tags: # don't bother comparing if we know this atom changes
                # If atom changed, add
                if atoms_are_different(prod_atoms[i], reac_atoms[j]):
                    changed_atoms.append(reac_atoms[j])
                    changed_atom_tags.append(reac_tag)
                    break
                # If reac_tag appears multiple times, add (need for stoichometry > 1)
                if prod_atom_tags.count(reac_tag) > 1:
                    changed_atoms.append(reac_atoms[j])
                    changed_atom_tags.append(reac_tag)
                    break

    # Reactant atoms that do not appear in product (tagged leaving groups)
    for j, reac_tag in enumerate(reac_atom_tags):
        if reac_tag not in changed_atom_tags:
            if reac_tag not in prod_atom_tags:
                changed_atoms.append(reac_atoms[j])
                changed_atom_tags.append(reac_tag)

    # Atoms that change CHIRALITY (just tetrahedral for now...)
    tetra_atoms = get_tetrahedral_atoms(reactants, products)
    if VERBOSE:
        print('Found {} atom-mapped tetrahedral atoms that have chirality specified at least partially'.format(len(tetra_atoms)))
    [set_isotope_to_equal_mapnum(reactant) for reactant in reactants]
    [set_isotope_to_equal_mapnum(product) for product in products]
    for (atom_tag, ar, ap) in tetra_atoms:
        if VERBOSE: 
            print('For atom tag {}'.format(atom_tag))
            print('    reactant: {}'.format(ar.GetChiralTag()))
            print('    product:  {}'.format(ap.GetChiralTag()))
        if atom_tag in changed_atom_tags:
            if VERBOSE:
                print('-> atoms have changed (by more than just chirality!)')
        else:
            unchanged = check_tetrahedral_centers_equivalent(ar, ap) and \
                    ChiralType.CHI_UNSPECIFIED not in [ar.GetChiralTag(), ap.GetChiralTag()]
            if unchanged:
                if VERBOSE: 
                    print('-> atoms confirmed to have same chirality, no change')
            else:
                if VERBOSE:
                    print('-> atom changed chirality!!')
                # Make sure chiral change is next to the reaction center and not
                # a random specifidation (must be CONNECTED to a changed atom)
                tetra_adj_to_rxn = False
                for neighbor in ap.GetNeighbors():
                    if neighbor.HasProp('molAtomMapNumber'):
                        if neighbor.GetProp('molAtomMapNumber') in changed_atom_tags:
                            tetra_adj_to_rxn = True
                            break
                if tetra_adj_to_rxn:
                    if VERBOSE:
                        print('-> atom adj to reaction center, now included')
                    changed_atom_tags.append(atom_tag)
                    changed_atoms.append(ar)
                else:
                    if VERBOSE:
                        print('-> adj far from reaction center, not including')
    [clear_isotope(reactant) for reactant in reactants]
    [clear_isotope(product) for product in products]


    if VERBOSE: 
        print('{} tagged atoms in reactants change 1-atom properties'.format(len(changed_atom_tags)))
        for smarts in [atom.GetSmarts() for atom in changed_atoms]:
            print('  {}'.format(smarts))

    return changed_atoms, changed_atom_tags, err

def get_special_groups(mol):
    '''Given an RDKit molecule, this function returns a list of tuples, where
    each tuple contains the AtomIdx's for a special group of atoms which should 
    be included in a fragment all together. This should only be done for the 
    reactants, otherwise the products might end up with mapping mismatches

    We draw a distinction between atoms in groups that trigger that whole
    group to be included, and "unimportant" atoms in the groups that will not
    be included if another atom matches.'''

    # Define templates
    group_templates = [ 
        (range(3), '[OH0,SH0]=C[O,Cl,I,Br,F]',), # carboxylic acid / halogen
        (range(3), '[OH0,SH0]=CN',), # amide/sulfamide
        (range(4), 'S(O)(O)[Cl]',), # sulfonyl chloride
        (range(3), 'B(O)O',), # boronic acid/ester
        ((0,), '[Si](C)(C)C'), # trialkyl silane
        ((0,), '[Si](OC)(OC)(OC)'), # trialkoxy silane, default to methyl
        (range(3), '[N;H0;$(N-[#6]);D2]-,=[N;D2]-,=[N;D1]',), # azide
        (range(8), 'O=C1N([Br,I,F,Cl])C(=O)CC1',), # NBS brominating agent
        (range(11), 'Cc1ccc(S(=O)(=O)O)cc1'), # Tosyl
        ((7,), 'CC(C)(C)OC(=O)[N]'), # N(boc)
        ((4,), '[CH3][CH0]([CH3])([CH3])O'), # 
        (range(2), '[C,N]=[C,N]',), # alkene/imine
        (range(2), '[C,N]#[C,N]',), # alkyne/nitrile
        ((2,), 'C=C-[*]',), # adj to alkene
        ((2,), 'C#C-[*]',), # adj to alkyne
        ((2,), 'O=C-[*]',), # adj to carbonyl
        ((3,), 'O=C([CH3])-[*]'), # adj to methyl ketone
        ((3,), 'O=C([O,N])-[*]',), # adj to carboxylic acid/amide/ester
        (range(4), 'ClS(Cl)=O',), # thionyl chloride
        (range(2), '[Mg,Li,Zn,Sn][Br,Cl,I,F]',), # grinard/metal (non-disassociated)
        (range(3), 'S(O)(O)',), # SO2 group
        (range(2), 'N~N',), # diazo
        ((1,), '[!#6;R]@[#6;R]',), # adjacency to heteroatom in ring
        ((2,), '[a!c]:a:a',), # two-steps away from heteroatom in aromatic ring
        #((1,), 'c(-,=[*]):c([Cl,I,Br,F])',), # ortho to halogen on ring - too specific?
        #((1,), 'c(-,=[*]):c:c([Cl,I,Br,F])',), # meta to halogen on ring - too specific?
        ((0,), '[B,C](F)(F)F'), # CF3, BF3 should have the F3 included
    ]

    # Stereo-specific ones (where we will need to include neighbors)
    # Tetrahedral centers should already be okay...
    group_templates += [
        ((1,2,), '[*]/[CH]=[CH]/[*]'), # trans with two hydrogens
        ((1,2,), '[*]/[CH]=[CH]\[*]'), # cis with two hydrogens
        ((1,2,), '[*]/[CH]=[CH0]([*])\[*]'), # trans with one hydrogens
        ((1,2,), '[*]/[D3;H1]=[!D1]'), # specified on one end, can be N or C
    ]

    # Build list
    groups = []
    for (add_if_match, template) in group_templates:
        matches = mol.GetSubstructMatches(Chem.MolFromSmarts(template), useChirality=True)
        for match in matches:
            add_if = []
            for pattern_idx, atom_idx in enumerate(match):
                if pattern_idx in add_if_match:
                    add_if.append(atom_idx)
            groups.append((add_if, match))
    return groups

def expand_atoms_to_use(mol, atoms_to_use, groups=[], symbol_replacements=[]):
    '''Given an RDKit molecule and a list of AtomIdX which should be included
    in the reaction, this function expands the list of AtomIdXs to include one 
    nearest neighbor with special consideration of (a) unimportant neighbors and
    (b) important functional groupings'''

    # Copy
    new_atoms_to_use = atoms_to_use[:]

    # Look for all atoms in the current list of atoms to use
    for atom in mol.GetAtoms():
        if atom.GetIdx() not in atoms_to_use: continue
        # Ensure membership of changed atom is checked against group
        for group in groups:
            if int(atom.GetIdx()) in group[0]:
                if VERBOSE: 
                    print('adding group due to match')
                    try:
                        print('Match from molAtomMapNum {}'.format(
                            atom.GetProp('molAtomMapNumber'),
                        ))
                    except KeyError:
                        pass
                for idx in group[1]:
                    if idx not in atoms_to_use:
                        new_atoms_to_use.append(idx)
                        symbol_replacements.append((idx, convert_atom_to_wildcard(mol.GetAtomWithIdx(idx))))
        # Look for all nearest neighbors of the currently-included atoms
        for neighbor in atom.GetNeighbors():
            # Evaluate nearest neighbor atom to determine what should be included
            new_atoms_to_use, symbol_replacements = \
                    expand_atoms_to_use_atom(mol, new_atoms_to_use, neighbor.GetIdx(), 
                        groups=groups, symbol_replacements=symbol_replacements)
            
    return new_atoms_to_use, symbol_replacements

def expand_atoms_to_use_atom(mol, atoms_to_use, atom_idx, groups=[], symbol_replacements=[]):
    '''Given an RDKit molecule and a list of AtomIdx which should be included
    in the reaction, this function extends the list of atoms_to_use by considering 
    a candidate atom extension, atom_idx'''

    # See if this atom belongs to any special groups (highest priority)
    found_in_group = False
    for group in groups: # first index is atom IDs for match, second is what to include
        if int(atom_idx) in group[0]: # int correction
            if VERBOSE: 
                print('adding group due to match')
                try:
                    print('Match from molAtomMapNum {}'.format(
                        mol.GetAtomWithIdx(atom_idx).GetProp('molAtomMapNumber'),
                    ))
                except KeyError:
                    pass
            # Add the whole list, redundancies don't matter 
            # *but* still call convert_atom_to_wildcard!
            for idx in group[1]:
                if idx not in atoms_to_use:
                        atoms_to_use.append(idx)
                        symbol_replacements.append((idx, convert_atom_to_wildcard(mol.GetAtomWithIdx(idx))))
            found_in_group = True
    if found_in_group:  
        return atoms_to_use, symbol_replacements
        
    # How do we add an atom that wasn't in an identified important functional group?
    # Develop generalized SMARTS symbol

    # Skip current candidate atom if it is already included
    if atom_idx in atoms_to_use:
        return atoms_to_use, symbol_replacements

    # Include this atom
    atoms_to_use.append(atom_idx)

    # Look for suitable SMARTS replacement
    symbol_replacements.append((atom_idx, convert_atom_to_wildcard(mol.GetAtomWithIdx(atom_idx))))

    return atoms_to_use, symbol_replacements

def convert_atom_to_wildcard(atom):
    '''This function takes an RDKit atom and turns it into a wildcard 
    using heuristic generalization rules. This function should be used
    when candidate atoms are used to extend the reaction core for higher
    generalizability'''

    # Is this a terminal atom? We can tell if the degree is one
    if atom.GetDegree() == 1:
        symbol = '[' + atom.GetSymbol() + ';D1;H{}'.format(atom.GetTotalNumHs())
        if atom.GetFormalCharge() != 0:
            charges = re.search('([-+]+[1-9]?)', atom.GetSmarts())
            symbol = symbol.replace(';D1', ';{};D1'.format(charges.group()))
    
    else:
        # Initialize
        symbol = '['

        # Add atom primitive - atomic num and aromaticity (don't use COMPLETE wildcards)
        if atom.GetAtomicNum() != 6:
            symbol += '#{};'.format(atom.GetAtomicNum())
            if atom.GetIsAromatic():
                symbol += 'a;'
        elif atom.GetIsAromatic():
            symbol += 'c;'
        else:
            symbol += 'C;'

        # Charge is important
        if atom.GetFormalCharge() != 0:
            charges = re.search('([-+]+[1-9]?)', atom.GetSmarts())
            if charges: symbol += charges.group() + ';'

        # Strip extra semicolon
        if symbol[-1] == ';': symbol = symbol[:-1]

    # Close with label or with bracket
    label = re.search('\:[0-9]+\]', atom.GetSmarts())
    if label: 
        symbol += label.group()
    else:
        symbol += ']'

    if VERBOSE: 
        if symbol != atom.GetSmarts():
            print('Improved generality of atom SMARTS {} -> {}'.format(atom.GetSmarts(), symbol))

    return symbol

def reassign_atom_mapping(transform):
    '''This function takes an atom-mapped reaction SMILES and reassigns 
    the atom-mapping labels (numbers) from left to right, once 
    that transform has been canonicalized.'''

    all_labels = re.findall('\:([0-9]+)\]', transform)

    # Define list of replacements which matches all_labels *IN ORDER*
    replacements = []
    replacement_dict = {}
    counter = 1
    for label in all_labels: # keep in order! this is important
        if label not in replacement_dict:
            replacement_dict[label] = str(counter)
            counter += 1
        replacements.append(replacement_dict[label])

    # Perform replacements in order
    transform_newmaps = re.sub('\:[0-9]+\]', 
        lambda match: (':' + replacements.pop(0) + ']'),
        transform)

    return transform_newmaps

def get_strict_smarts_for_atom(atom):
    '''
    For an RDkit atom object, generate a SMARTS pattern that
    matches the atom as strictly as possible
    '''

    symbol = atom.GetSmarts()
    if atom.GetSymbol() == 'H':
        symbol = '[#1]'

    if '[' not in symbol:
        symbol = '[' + symbol + ']'

    # Explicit stereochemistry - *before* H
    if USE_STEREOCHEMISTRY:
        if atom.GetChiralTag() != Chem.rdchem.ChiralType.CHI_UNSPECIFIED:
            if '@' not in symbol:
                # Be explicit when there is a tetrahedral chiral tag
                if atom.GetChiralTag() == Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW:
                    tag = '@'
                elif atom.GetChiralTag() == Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW:
                    tag = '@@'
                if ':' in symbol:
                    symbol = symbol.replace(':', ';{}:'.format(tag))
                else:
                    symbol = symbol.replace(']', ';{}]'.format(tag))

    if 'H' not in symbol:
        H_symbol = 'H{}'.format(atom.GetTotalNumHs())
        # Explicit number of hydrogens: include "H0" when no hydrogens present
        if ':' in symbol: # stick H0 before label
            symbol = symbol.replace(':', ';{}:'.format(H_symbol))
        else:
            symbol = symbol.replace(']', ';{}]'.format(H_symbol))
      
    # Explicit degree
    if ':' in symbol:
        symbol = symbol.replace(':', ';D{}:'.format(atom.GetDegree()))
    else:
        symbol = symbol.replace(']', ';D{}]'.format(atom.GetDegree()))

    # Explicit formal charge
    if '+' not in symbol and '-' not in symbol:
        charge = atom.GetFormalCharge()
        charge_symbol = '+' if (charge >= 0) else '-'
        charge_symbol += '{}'.format(abs(charge))
        if ':' in symbol: 
            symbol = symbol.replace(':', ';{}:'.format(charge_symbol))
        else:
            symbol = symbol.replace(']', ';{}]'.format(charge_symbol))

    return symbol

def expand_changed_atom_tags(changed_atom_tags, reactant_fragments):
    '''Given a list of changed atom tags (numbers as strings) and a string consisting
    of the reactant_fragments to include in the reaction transform, this function 
    adds any tagged atoms found in the reactant side of the template to the 
    changed_atom_tags list so that those tagged atoms are included in the products'''

    expansion = []
    atom_tags_in_reactant_fragments = re.findall('\:([[0-9]+)\]', reactant_fragments)
    for atom_tag in atom_tags_in_reactant_fragments:
        if atom_tag not in changed_atom_tags:
            expansion.append(atom_tag)
    if VERBOSE: print('after building reactant fragments, additional labels included: {}'.format(expansion))
    return expansion

def get_fragments_for_changed_atoms(mols, changed_atom_tags, radius=0, 
    category='reactants', expansion=[]):
    '''Given a list of RDKit mols and a list of changed atom tags, this function
    computes the SMILES string of molecular fragments using MolFragmentToSmiles 
    for all changed fragments.

    expansion: atoms added during reactant expansion that should be included and
               generalized in product fragment
    '''
    fragments = ''
    mols_changed = []
    for mol in mols:
        # Initialize list of replacement symbols (updated during expansion)
        symbol_replacements = []

        # Are we looking for special reactive groups? (reactants only)
        if category == 'reactants':
            groups = get_special_groups(mol)
        else:
            groups = []

        # Build list of atoms to use
        atoms_to_use = []
        for atom in mol.GetAtoms():
            # Check self (only tagged atoms)
            if ':' in atom.GetSmarts():
                if atom.GetSmarts().split(':')[1][:-1] in changed_atom_tags:
                    atoms_to_use.append(atom.GetIdx())
                    symbol = get_strict_smarts_for_atom(atom)
                    if symbol != atom.GetSmarts():
                        symbol_replacements.append((atom.GetIdx(), symbol))
                    continue

        # Fully define leaving groups and this molecule participates?
        if INCLUDE_ALL_UNMAPPED_REACTANT_ATOMS and len(atoms_to_use) > 0:
            if category == 'reactants':
                for atom in mol.GetAtoms():
                    if not atom.HasProp('molAtomMapNumber'):
                        atoms_to_use.append(atom.GetIdx())

        # Check neighbors (any atom)
        for k in range(radius):
            atoms_to_use, symbol_replacements = expand_atoms_to_use(mol, atoms_to_use, 
                groups=groups, symbol_replacements=symbol_replacements)

        if category == 'products':
            # Add extra labels to include (for products only)
            if expansion:
                for atom in mol.GetAtoms():
                    if ':' not in atom.GetSmarts(): continue
                    label = atom.GetSmarts().split(':')[1][:-1]
                    if label in expansion and label not in changed_atom_tags:
                        atoms_to_use.append(atom.GetIdx())
                        # Make the expansion a wildcard
                        symbol_replacements.append((atom.GetIdx(), convert_atom_to_wildcard(atom))) 
                        if VERBOSE: print('expanded label {} to wildcard in products'.format(label))
            
            # Make sure unmapped atoms are included (from products)
            for atom in mol.GetAtoms():
                if not atom.HasProp('molAtomMapNumber'): 
                    atoms_to_use.append(atom.GetIdx())
                    symbol = get_strict_smarts_for_atom(atom)
                    symbol_replacements.append((atom.GetIdx(), symbol))

        # Define new symbols based on symbol_replacements
        symbols = [atom.GetSmarts() for atom in mol.GetAtoms()]
        for (i, symbol) in symbol_replacements:
            symbols[i] = symbol

        if not atoms_to_use: 
            continue
        
        # Keep flipping stereocenters until we are happy...
        # this is a sloppy fix during extraction to achieve consistency
        tetra_consistent = False
        num_tetra_flips = 0
        while not tetra_consistent and num_tetra_flips < 100:
            mol_copy = deepcopy(mol)
            [x.ClearProp('molAtomMapNumber') for x in mol_copy.GetAtoms()]   
            this_fragment = AllChem.MolFragmentToSmiles(mol_copy, atoms_to_use, 
                atomSymbols=symbols, allHsExplicit=True, 
                isomericSmiles=USE_STEREOCHEMISTRY, allBondsExplicit=True)

            # Figure out what atom maps are tetrahedral centers
            # Set isotopes to make sure we're getting the *exact* match we want
            this_fragment_mol = AllChem.MolFromSmarts(this_fragment)
            tetra_map_nums = []
            for atom in this_fragment_mol.GetAtoms():
                if atom.HasProp('molAtomMapNumber'):
                    atom.SetIsotope(int(atom.GetProp('molAtomMapNumber')))
                    if atom.GetChiralTag() != Chem.rdchem.ChiralType.CHI_UNSPECIFIED:
                        tetra_map_nums.append(atom.GetProp('molAtomMapNumber'))
            map_to_id = {}
            for atom in mol.GetAtoms():
                if atom.HasProp('molAtomMapNumber'):
                    atom.SetIsotope(int(atom.GetProp('molAtomMapNumber')))
                    map_to_id[atom.GetProp('molAtomMapNumber')] = atom.GetIdx()
            
            # Look for matches
            tetra_consistent = True
            all_matched_ids = []
            
            # skip substructure matching if there are a lot of fragments
            # this can help prevent GetSubstructMatches from hanging 
            try:
                frag_smi = Chem.MolToSmiles(this_fragment_mol)
                if frag_smi.count('.') > 5:
                    break
            except Exception as e:
                print(e)
                break
            
            for matched_ids in mol.GetSubstructMatches(this_fragment_mol, useChirality=True):
                all_matched_ids.extend(matched_ids)
            shuffle(tetra_map_nums)
            for tetra_map_num in tetra_map_nums:
                if VERBOSE: print('Checking consistency of tetrahedral {}'.format(tetra_map_num))
                #print('Using fragment {}'.format(Chem.MolToSmarts(this_fragment_mol, True)))
                if map_to_id[tetra_map_num] not in all_matched_ids:
                    tetra_consistent = False
                    if VERBOSE: print('@@@@@@@@@@@ FRAGMENT DOES NOT MATCH PARENT MOL @@@@@@@@@@@@@@')
                    if VERBOSE: print('@@@@@@@@@@@ FLIPPING CHIRALITY SYMBOL NOW      @@@@@@@@@@@@@@')
                    prevsymbol = symbols[map_to_id[tetra_map_num]]
                    if '@@' in prevsymbol:
                        symbol = prevsymbol.replace('@@', '@')
                    elif '@' in prevsymbol:
                        symbol = prevsymbol.replace('@', '@@')
                    else:
                        raise ValueError('Need to modify symbol of tetra atom without @ or @@??')
                    symbols[map_to_id[tetra_map_num]] = symbol
                    num_tetra_flips += 1
                    # IMPORTANT: only flip one at a time
                    break 

            # Clear isotopes
            for atom in mol.GetAtoms():
                atom.SetIsotope(0)
        
        if not tetra_consistent:
            raise ValueError('Could not find consistent tetrahedral mapping, {} centers'.format(len(tetra_map_nums)))

        fragments += '(' + this_fragment + ').'
        mols_changed.append(Chem.MolToSmiles(clear_mapnum(Chem.MolFromSmiles(Chem.MolToSmiles(mol, True))), True))

    # auxiliary template information: is this an intramolecular reaction or dimerization?
    intra_only = (1 == len(mols_changed))
    dimer_only = (1 == len(set(mols_changed))) and (len(mols_changed) == 2)
    
    return fragments[:-1], intra_only, dimer_only

def canonicalize_transform(transform):
    '''This function takes an atom-mapped SMARTS transform and
    converts it to a canonical form by, if nececssary, rearranging
    the order of reactant and product templates and reassigning
    atom maps.'''

    transform_reordered = '>>'.join([canonicalize_template(x) for x in transform.split('>>')])
    return reassign_atom_mapping(transform_reordered)

def canonicalize_template(template):
    '''This function takes one-half of a template SMARTS string 
    (i.e., reactants or products) and re-orders them based on
    an equivalent string without atom mapping.'''

    # Strip labels to get sort orders
    template_nolabels = re.sub('\:[0-9]+\]', ']', template)

    # Split into separate molecules *WITHOUT wrapper parentheses*
    template_nolabels_mols = template_nolabels[1:-1].split(').(')
    template_mols          = template[1:-1].split(').(')

    # Split into fragments within those molecules
    for i in range(len(template_mols)):
        nolabel_mol_frags = template_nolabels_mols[i].split('.')
        mol_frags         = template_mols[i].split('.')

        # Get sort order within molecule, defined WITHOUT labels
        sortorder = [j[0] for j in sorted(enumerate(nolabel_mol_frags), key = lambda x:x[1])]

        # Apply sorting and merge list back into overall mol fragment
        template_nolabels_mols[i] = '.'.join([nolabel_mol_frags[j] for j in sortorder])
        template_mols[i]          = '.'.join([mol_frags[j] for j in sortorder])

    # Get sort order between molecules, defined WITHOUT labels
    sortorder = [j[0] for j in sorted(enumerate(template_nolabels_mols), key = lambda x:x[1])]

    # Apply sorting and merge list back into overall transform
    template = '(' + ').('.join([template_mols[i] for i in sortorder]) + ')'

    return template

def bond_to_label(bond):
    '''This function takes an RDKit bond and creates a label describing
    the most important attributes'''
    a1_label = str(bond.GetBeginAtom().GetAtomicNum())
    a2_label = str(bond.GetEndAtom().GetAtomicNum())
    if bond.GetBeginAtom().HasProp('molAtomMapNumber'):
        a1_label += bond.GetBeginAtom().GetProp('molAtomMapNumber')
    if bond.GetEndAtom().HasProp('molAtomMapNumber'):
        a2_label += bond.GetEndAtom().GetProp('molAtomMapNumber')
    atoms = sorted([a1_label, a2_label])

    return '{}{}{}'.format(atoms[0], bond.GetSmarts(), atoms[1])

def extract_from_reaction(reaction):
    reactants = mols_from_smiles_list(replace_deuterated(reaction['reactants']).split('.'))
    products = mols_from_smiles_list(replace_deuterated(reaction['products']).split('.'))
    
    # if rdkit cant understand molecule, return
    if None in reactants: 
        return {'reaction_id': reaction['_id'], 'err_msg': 'rdkit cant understand reactants molecule'}
    if None in products: 
        print(reaction['products'])
        return {'reaction_id': reaction['_id'], 'err_msg': 'rdkit cant understand products molecule'}
    
    # try to sanitize molecules
    try:
        for i in range(len(reactants)):
            reactants[i] = AllChem.RemoveHs(reactants[i]) # *might* not be safe
        for i in range(len(products)):
            products[i] = AllChem.RemoveHs(products[i]) # *might* not be safe
        [Chem.SanitizeMol(mol) for mol in reactants + products] # redundant w/ RemoveHs
        [mol.UpdatePropertyCache() for mol in reactants + products]
    except Exception as e:
        # can't sanitize -> skip
        err_msg = str(e) + ', Could not load SMILES or sanitize'
        print(err_msg)
        print('ID: {}'.format(reaction['_id']))
        return {'reaction_id': reaction['_id'], 'err_msg': err_msg}
    
    are_unmapped_product_atoms = False
    extra_reactant_fragment = ''
    for product in products:
        prod_atoms = product.GetAtoms()
        if sum([a.HasProp('molAtomMapNumber') for a in prod_atoms]) < len(prod_atoms):
            if VERBOSE: print('Not all product atoms have atom mapping')
            if VERBOSE: print('ID: {}'.format(reaction['_id']))
            are_unmapped_product_atoms = True

    if are_unmapped_product_atoms: # add fragment to template
        for product in products:
            # Get unmapped atoms
            unmapped_ids = [
                a.GetIdx() for a in prod_atoms if not a.HasProp('molAtomMapNumber')
            ]
            if len(unmapped_ids) > MAXIMUM_NUMBER_UNMAPPED_PRODUCT_ATOMS:
                # Skip this example - too many unmapped product atoms!
                return
            # Define new atom symbols for fragment with atom maps, generalizing fully
            atom_symbols = ['[{}]'.format(a.GetSymbol()) for a in prod_atoms]
            # And bond symbols...
            bond_symbols = ['~' for b in product.GetBonds()]
            if unmapped_ids:
                extra_reactant_fragment += AllChem.MolFragmentToSmiles(
                    product, unmapped_ids, 
                    allHsExplicit = False, isomericSmiles = USE_STEREOCHEMISTRY, 
                    atomSymbols = atom_symbols, bondSymbols = bond_symbols
                ) + '.'
        if extra_reactant_fragment:
            extra_reactant_fragment = extra_reactant_fragment[:-1]
            if VERBOSE: print('    extra reactant fragment: {}'.format(extra_reactant_fragment))

        # Consolidate repeated fragments (stoichometry)
        extra_reactant_fragment = '.'.join(sorted(list(set(extra_reactant_fragment.split('.')))))


    if None in reactants + products:
        err_msg = 'Could not parse all molecules in reaction, skipping'
        print(err_msg)
        print('ID: {}'.format(reaction['_id']))
        return {'reaction_id': reaction['_id'], 'err_msg': err_msg}

    # Calculate changed atoms
    changed_atoms, changed_atom_tags, err = get_changed_atoms(reactants, products)
    if err:
        err_msg = 'Could not get changed atoms'
        if VERBOSE:
            print(err_msg)
            print('ID: {}'.format(reaction['_id']))
        return {'reaction_id': reaction['_id'], 'err_msg': err_msg}
    if not changed_atom_tags:
        err_msg = 'No atoms changed?'
        if VERBOSE:
            print(err_msg)
            print('ID: {}'.format(reaction['_id']))
        # print('Reaction SMILES: {}'.format(example_doc['RXN_SMILES']))
        return {'reaction_id': reaction['_id'], 'err_msg': err_msg}

    try:
        # Get fragments for reactants
        reactant_fragments, intra_only, dimer_only = get_fragments_for_changed_atoms(reactants, changed_atom_tags, 
            radius = 1, expansion = [], category = 'reactants')
        # Get fragments for products 
        # (WITHOUT matching groups but WITH the addition of reactant fragments)
        product_fragments, _, _  = get_fragments_for_changed_atoms(products, changed_atom_tags, 
            radius = 0, expansion = expand_changed_atom_tags(changed_atom_tags, reactant_fragments),
            category = 'products')
    except ValueError as e:
        err_msg = str(e)
        if VERBOSE:
            print(err_msg)
            print(reaction['_id'])
        return {'reaction_id': reaction['_id'], 'err_msg': err_msg}

    # Put together and canonicalize (as best as possible)
    rxn_string = '{}>>{}'.format(reactant_fragments, product_fragments)
    rxn_canonical = canonicalize_transform(rxn_string)
    # Change from inter-molecular to intra-molecular 
    rxn_canonical_split = rxn_canonical.split('>>')
    rxn_canonical = rxn_canonical_split[0][1:-1].replace(').(', '.') + \
        '>>' + rxn_canonical_split[1][1:-1].replace(').(', '.')

    reactants_string = rxn_canonical.split('>>')[0]
    products_string  = rxn_canonical.split('>>')[1]

    retro_canonical = products_string + '>>' + reactants_string

    # Load into RDKit
    rxn = AllChem.ReactionFromSmarts(retro_canonical)
    if rxn.Validate()[1] != 0: 
        err_msg = 'Could not validate reaction successfully' + ', retro_canonical: {}'.format(retro_canonical)
        print(err_msg)
        print('ID: {}'.format(reaction['_id']))
        if VERBOSE: raw_input('Pausing...')
        return {'reaction_id': reaction['_id'], 'err_msg': err_msg}

    template = {
        'products': products_string,
        'reactants': reactants_string,
        'reaction_smarts': retro_canonical,
        'intra_only': intra_only,
        'dimer_only': dimer_only,
        'reaction_id': reaction['_id'],
        'necessary_reagent': extra_reactant_fragment,
    }
    
    return template



import networkx as nx
from g2s.utils.chem_utils import BOND_TYPES
from rdkit import Chem
from typing import List, Tuple, Union


def get_sub_mol(mol, sub_atoms):
    new_mol = Chem.RWMol()
    atom_map = {}
    for idx in sub_atoms:
        atom = mol.GetAtomWithIdx(idx)
        atom_map[idx] = new_mol.AddAtom(atom)

    sub_atoms = set(sub_atoms)
    for idx in sub_atoms:
        a = mol.GetAtomWithIdx(idx)
        for b in a.GetNeighbors():
            if b.GetIdx() not in sub_atoms:
                continue
            bond = mol.GetBondBetweenAtoms(a.GetIdx(), b.GetIdx())
            bt = bond.GetBondType()
            if a.GetIdx() < b.GetIdx():  # each bond is enumerated twice
                new_mol.AddBond(atom_map[a.GetIdx()], atom_map[b.GetIdx()], bt)

    return new_mol.GetMol()


class RxnGraph:
    """
    RxnGraph is an abstract class for storing all elements of a reaction, like
    reactants, products and fragments. The edits associated with the reaction
    are also captured in edit labels. One can also use h_labels, which keep track
    of atoms with hydrogen changes. For reactions with multiple edits, a done
    label is also added to account for termination of edits.
    """

    def __init__(self,
                 prod_mol: Chem.Mol = None,
                 frag_mol: Chem.Mol = None,
                 reac_mol: Chem.Mol = None,
                 rxn_class: int = None) -> None:
        """
        Parameters
        ----------
        prod_mol: Chem.Mol,
            Product molecule
        frag_mol: Chem.Mol, default None
            Fragment molecule(s)
        reac_mol: Chem.Mol, default None
            Reactant molecule(s)
        rxn_class: int, default None,
            Reaction class for this reaction.
        """
        if prod_mol is not None:
            self.prod_mol = RxnElement(mol=prod_mol, rxn_class=rxn_class)
        if frag_mol is not None:
            self.frag_mol = MultiElement(mol=frag_mol, rxn_class=rxn_class)
        if reac_mol is not None:
            self.reac_mol = MultiElement(mol=reac_mol, rxn_class=rxn_class)
        self.rxn_class = rxn_class

    def get_attributes(self, mol_attrs: Tuple = ('prod_mol', 'frag_mol', 'reac_mol')) -> Tuple:
        """
        Returns the different attributes associated with the reaction graph.

        Parameters
        ----------
        mol_attrs: Tuple,
            Molecule objects to return
        """
        return tuple(getattr(self, attr) for attr in mol_attrs if hasattr(self, attr))


class RxnElement:
    """
    RxnElement is an abstract class for dealing with single molecule. The graph
    and corresponding molecule attributes are built for the molecule. The constructor
    accepts only mol objects, sidestepping the use of SMILES string which may always
    not be achievable, especially for a unkekulizable molecule.
    """

    def __init__(self, mol: Chem.Mol, rxn_class: int = None) -> None:
        """
        Parameters
        ----------
        mol: Chem.Mol,
            Molecule
        rxn_class: int, default None,
            Reaction class for this reaction.
        """
        self.mol = mol
        self.rxn_class = rxn_class
        self._build_mol()
        self._build_graph()

    def _build_mol(self) -> None:
        """Builds the molecule attributes."""
        self.num_atoms = self.mol.GetNumAtoms()
        self.num_bonds = self.mol.GetNumBonds()
        self.amap_to_idx = {atom.GetAtomMapNum(): atom.GetIdx()
                            for atom in self.mol.GetAtoms()}
        self.idx_to_amap = {value: key for key, value in self.amap_to_idx.items()}

    def _build_graph(self) -> None:
        """Builds the graph attributes."""
        self.G_undir = nx.Graph(Chem.rdmolops.GetAdjacencyMatrix(self.mol))
        self.G_dir = nx.DiGraph(Chem.rdmolops.GetAdjacencyMatrix(self.mol))

        for atom in self.mol.GetAtoms():
            self.G_undir.nodes[atom.GetIdx()]['label'] = atom.GetSymbol()
            self.G_dir.nodes[atom.GetIdx()]['label'] = atom.GetSymbol()

        for bond in self.mol.GetBonds():
            a1 = bond.GetBeginAtom().GetIdx()
            a2 = bond.GetEndAtom().GetIdx()
            btype = BOND_TYPES.index(bond.GetBondType())
            self.G_undir[a1][a2]['label'] = btype
            self.G_dir[a1][a2]['label'] = btype
            self.G_dir[a2][a1]['label'] = btype

        self.atom_scope = (0, self.num_atoms)
        self.bond_scope = (0, self.num_bonds)

    def update_atom_scope(self, offset: int) -> Union[List, Tuple]:
        """Updates the atom indices by the offset.

        Parameters
        ----------
        offset: int,
            Offset to apply
        """
        # Note that the self. reference to atom_scope is dropped to keep self.atom_scope non-dynamic
        if isinstance(self.atom_scope, list):
            atom_scope = [(st + offset, le) for st, le in self.atom_scope]
        else:
            st, le = self.atom_scope
            atom_scope = (st + offset, le)

        return atom_scope

    def update_bond_scope(self, offset: int) -> Union[List, Tuple]:
        """Updates the bond indices by the offset.

        Parameters
        ----------
        offset: int,
            Offset to apply
        """
        # Note that the self. reference to bond_scope is dropped to keep self.bond_scope non-dynamic
        if isinstance(self.bond_scope, list):
            bond_scope = [(st + offset, le) for st, le in self.bond_scope]
        else:
            st, le = self.bond_scope
            bond_scope = (st + offset, le)

        return bond_scope


class MultiElement(RxnElement):
    """
    MultiElement is an abstract class for dealing with multiple molecules. The graph
    is built with all molecules, but different molecules and their sizes are stored.
    The constructor accepts only mol objects, sidestepping the use of SMILES string
    which may always not be achievable, especially for an invalid intermediates.
    """

    def _build_graph(self) -> None:
        """Builds the graph attributes."""
        self.G_undir = nx.Graph(Chem.rdmolops.GetAdjacencyMatrix(self.mol))
        self.G_dir = nx.DiGraph(Chem.rdmolops.GetAdjacencyMatrix(self.mol))

        for atom in self.mol.GetAtoms():
            self.G_undir.nodes[atom.GetIdx()]['label'] = atom.GetSymbol()
            self.G_dir.nodes[atom.GetIdx()]['label'] = atom.GetSymbol()

        for bond in self.mol.GetBonds():
            a1 = bond.GetBeginAtom().GetIdx()
            a2 = bond.GetEndAtom().GetIdx()
            btype = BOND_TYPES.index(bond.GetBondType())
            self.G_undir[a1][a2]['label'] = btype
            self.G_dir[a1][a2]['label'] = btype
            self.G_dir[a2][a1]['label'] = btype

        frag_indices = [c for c in nx.strongly_connected_components(self.G_dir)]
        self.mols = [get_sub_mol(self.mol, sub_atoms) for sub_atoms in frag_indices]

        atom_start = 0
        bond_start = 0
        self.atom_scope = []
        self.bond_scope = []

        for mol in self.mols:
            self.atom_scope.append((atom_start, mol.GetNumAtoms()))
            self.bond_scope.append((bond_start, mol.GetNumBonds()))
            atom_start += mol.GetNumAtoms()
            bond_start += mol.GetNumBonds()



import re
from rdkit import Chem
# from retroformer.rdchiral.template_extractor import get_changed_atoms, mols_from_smiles_list, \
#     replace_deuterated
from retro_star.rdchiral.template_extractor import get_changed_atoms, mols_from_smiles_list, \
    replace_deuterated

import numpy as np

BONDTYPES = ['NONE', 'AROMATIC', 'DOUBLE', 'SINGLE', 'TRIPLE']
BONDTOI = {bond: i for i, bond in enumerate(BONDTYPES)}


class SmilesGraph:
    def __init__(self, smi, existing=None):
        self.V = len(smi_tokenizer(smi))
        self.smi = smi
        if existing is None:
            self.adjacency_matrix, self.bond_type_dict, self.bond_attributes = self.extract_graph_structure(smi)
        else:
            assert len(existing) == 3
            self.adjacency_matrix, self.bond_type_dict, self.bond_attributes = existing

        self.full_adjacency_tensor = np.zeros((len(self.adjacency_matrix), len(self.adjacency_matrix), 7), dtype=int)
        for i in range(len(self.adjacency_matrix)):
            for j in self.adjacency_matrix[i]:
                self.full_adjacency_tensor[i][j] = self.bond_attributes[(i, j)]

    def one_hot_vector(self, val, lst):
        """Converts a value to a one-hot vector based on options in list"""
        if val not in lst:
            val = lst[-1]
        return map(lambda x: x == val, lst)

    def get_bond_attributes(self, bond):
        """
        From Neural FP defaults:
        The bond features were a concatenation of whether the bond type was single, double, triple,
        or aromatic, whether the bond was conjugated, and whether the bond was part of a ring.
        """
        # Initialize
        attributes = []
        # Add bond type
        attributes += self.one_hot_vector(
            bond.GetBondTypeAsDouble(),
            [1.0, 1.5, 2.0, 3.0]
        )
        # Add if is aromatic
        attributes.append(bond.GetIsAromatic())
        # Add if bond is conjugated
        attributes.append(bond.GetIsConjugated())
        # Add if bond is part of ring
        attributes.append(bond.IsInRing())

        return np.array(attributes)

    def extract_graph_structure(self, smi, verbose=False):
        """Build SMILES graph from molecule graph"""
        adjacency_matrix = [[] for _ in range(len(smi_tokenizer(smi)))]
        bond_types = {}
        bond_attributes = {}
        sample_mol = Chem.MolFromSmiles(smi)
        atom_ordering = [atom.GetIdx() for atom in sample_mol.GetAtoms()]
        atom_symbols = [atom.GetSmarts() for atom in sample_mol.GetAtoms()]
        neighbor_smiles_list = []
        neighbor_bonds_list, neighbor_bonds_attributes_list = [], []
        for atom in sample_mol.GetAtoms():
            neighbor_bonds = []
            neighbor_bonds_attributes = []
            atom_symbols_i = atom_symbols[:]
            atom_symbols_i[atom.GetIdx()] = '[{}:1]'.format(atom.GetSymbol())
            for i, neighbor_atom in enumerate(atom.GetNeighbors()):
                atom_symbols_i[neighbor_atom.GetIdx()] = '[{}:{}]'.format(neighbor_atom.GetSymbol(), 900 + i)
                bond = sample_mol.GetBondBetweenAtoms(atom.GetIdx(), neighbor_atom.GetIdx())
                neighbor_bonds.append(str(bond.GetBondType()))
                neighbor_bonds_attributes.append(self.get_bond_attributes(bond))

            neighbor_tagged_smiles = \
                Chem.MolFragmentToSmiles(sample_mol, atomsToUse=atom_ordering,
                                         canonical=False, atomSymbols=atom_symbols_i)
            neighbor_smiles_list.append(neighbor_tagged_smiles)
            neighbor_bonds_list.append(neighbor_bonds)
            neighbor_bonds_attributes_list.append(neighbor_bonds_attributes)

        for ni, neighbor_tagged_smiles in enumerate(neighbor_smiles_list):
            neighbor_tagged_tokens = smi_tokenizer(neighbor_tagged_smiles)
            neighbor_bonds = neighbor_bonds_list[ni]
            neighbor_bonds_attributes = neighbor_bonds_attributes_list[ni]
            cur_i, cand_js, order = -1, [], []
            for j in range(len(neighbor_tagged_tokens)):
                if re.match('\[.*:1\]', neighbor_tagged_tokens[j]):
                    cur_i = j
                if re.match('\[.*:90[0-9]\]', neighbor_tagged_tokens[j]):
                    cand_js.append(j)
                    order.append(int(re.match('\[.*:(90[0-9])\]', neighbor_tagged_tokens[j]).group(1)) - 900)
            if cur_i > -1:
                assert len(neighbor_bonds) == len(cand_js)
                neighbor_bonds = list(np.array(neighbor_bonds)[order])
                neighbor_bonds_attributes = list(np.array(neighbor_bonds_attributes)[order])
                if verbose:
                    print(neighbor_tagged_smiles)
                    print(cur_i, cand_js, neighbor_bonds, '\n')
                adjacency_matrix[cur_i] = cand_js
                for cur_j in cand_js:
                    bond_types[(cur_i, cur_j)] = BONDTOI[neighbor_bonds.pop(0)]
                    bond_attributes[(cur_i, cur_j)] = neighbor_bonds_attributes.pop(0)
        return adjacency_matrix, bond_types, bond_attributes


def set_distance(a, b):
    return np.min([len(set(a) - set(b)), len(set(b) - set(a))])


def set_overlap(a, b):
    return len(set(a).intersection(set(b)))


def select_diverse_candidate(cc_trace_with_score, diverse_k=10):
    """Select top-k diverse candidate greedily from a list of tuple of (list of nodes, scores)"""
    selected_indices = {0}
    selected_cc_trace_with_score = [cc_trace_with_score[0]]
    pair2distance = {}

    for _ in range(min(diverse_k, len(cc_trace_with_score)) - 1):
        distance = []
        explore_new = False
        for i in range(len(cc_trace_with_score)):
            if i in selected_indices:
                distance.append(-float('inf'))
                continue
            current_cc_trace, current_cc_score = cc_trace_with_score[i]
            # avg_distance, tmp_trace = [], None
            distance_min, trace_min = float('inf'), ()
            for selected_cc_trace, _ in selected_cc_trace_with_score:
                pair_key = tuple(sorted([selected_cc_trace, current_cc_trace]))
                if pair_key not in pair2distance:
                    pair2distance[pair_key] = set_distance(selected_cc_trace, current_cc_trace)

                if pair2distance[pair_key] < distance_min:
                    distance_min = pair2distance[pair_key]
                    trace_min = selected_cc_trace

            if (distance_min > 2 and set_overlap(trace_min, current_cc_trace) < 3) \
                    or ((distance_min == 2) and len(trace_min) == 2):
                distance.append(distance_min + current_cc_score)
                explore_new = True
            else:
                distance.append(-float('inf'))

        if not explore_new:
            break

        top_index = np.argsort(distance)[-1]
        selected_indices.add(top_index)
        selected_cc_trace_with_score.append(cc_trace_with_score[top_index])

    return selected_cc_trace_with_score


def get_reaction_centers_from_template(src_smiles, blank_src_smiles, graph_pack, reaction_centers):
    """Retrieve all the potential reaction center from a pool of existing molecule fragment"""
    mol_blank = Chem.MolFromSmiles(blank_src_smiles)
    mol = Chem.MolFromSmiles(src_smiles)
    if mol is None:
        return []
    potential_rcs = {}
    for rc in (reaction_centers):
        if rc[0] == '(' and rc[-1] == ')':
            rc = rc[1:-1]
        patt = Chem.MolFromSmarts(rc)
        if patt is not None and mol_blank.HasSubstructMatch(patt):
            for match in mol_blank.GetSubstructMatches(patt):
                token_match_indices = []
                for index in match:
                    atom_smarts = mol.GetAtomWithIdx(index).GetSmarts()
                    token_match_indices.append(int(re.match('.*:([0-9]+)\]', atom_smarts).group(1)))
                if tuple(sorted(token_match_indices)) not in potential_rcs:
                    score = get_cc_score(token_match_indices, graph_pack) / get_norm(token_match_indices, graph_pack)
                    potential_rcs[tuple(sorted(token_match_indices))] = score

    return sorted(potential_rcs.items(), key=lambda x: -x[1])


def dfs_cc(trace, i, visited, graph_pack, alpha_atom=0.01, alpha_bond=0.01):
    """Retrieve the connected components considering both the atom and bond reactive probability"""
    node_scores, edge_scores, adjacency_matrix, _ = graph_pack
    visited[i] = True
    neighbors = [j for j in adjacency_matrix[i] if not visited[j] and
                 node_scores[j] > alpha_atom and (edge_scores[i, j] > alpha_bond or edge_scores[j, i] > alpha_bond)]
    if not neighbors:
        return trace
    for j in neighbors:
        if not visited[j]:
            trace = dfs_cc(trace + [j], j, visited, graph_pack, alpha_atom, alpha_bond)
    return trace


def dfs_cc_atom(trace, i, visited, graph_pack, alpha_atom=0.01):
    """Retrieve the connected components considering only the atom reactive probability"""
    node_scores, edge_scores, adjacency_matrix, _ = graph_pack
    visited[i] = True
    neighbors = [j for j in adjacency_matrix[i] if not visited[j] and node_scores[j] > alpha_atom]
    if not neighbors:
        return trace
    for j in neighbors:
        if not visited[j]:
            trace = dfs_cc_atom(trace + [j], j, visited, graph_pack, alpha_atom)
    return trace


def dfs_cc_bond(trace, i, visited, graph_pack, cc_trace_parent, alpha_bond=0.01):
    """Retrieve the connected components considering only the bond reactive probability"""
    node_scores, edge_scores, adjacency_matrix, _ = graph_pack
    visited[i] = True
    neighbors = [j for j in adjacency_matrix[i] if not visited[j] and
                 (edge_scores[i, j] > alpha_bond and edge_scores[j, i] > alpha_bond) and
                 j in cc_trace_parent]

    if not neighbors:
        return trace
    for j in neighbors:
        if not visited[j]:
            trace = dfs_cc_bond(trace + [j], j, visited, graph_pack, cc_trace_parent, alpha_bond)
    return trace


def get_cc_score(cc_trace, graph_pack):
    """Retrieve the total reactive scores given a subgraph"""
    node_scores, edge_scores, adjacency_matrix, full_adjacency_matrix = graph_pack
    sub_edge_scores = edge_scores[list(cc_trace)][:, list(cc_trace)]
    sub_adj_matrix = full_adjacency_matrix[list(cc_trace)][:, list(cc_trace)]
    return sum(np.log(node_scores[list(cc_trace)])) + sum(np.log(sub_edge_scores[sub_adj_matrix > 0]))


def get_norm(cc_trace, graph_pack):
    """Retrieve the normalization factor for the normalized reactive scores"""
    node_scores, edge_scores, adjacency_matrix, full_adjacency_matrix = graph_pack
    num_nodes = len(cc_trace)
    num_edges = np.sum(full_adjacency_matrix[list(cc_trace)][:, list(cc_trace)] > 0)

    return num_nodes + num_edges


def get_boarder_cycles(cc_trace, full_adjacency_matrix):
    """Identify which nodes are within the border cycle"""
    def dfs_fc(i, trace, adj_matrix):
        neighbors = np.argwhere(adj_matrix[i] > 0).flatten()
        for j in neighbors:
            if j in trace and len(trace) - trace.index(j) > 2:
                # find cycle
                cycle = sorted(trace[trace.index(j):])
                cycles.add(tuple(cycle))
                return
            elif j not in trace:
                dfs_fc(j, trace + [j], adj_matrix)

    cycles, boarder_cycles, boarder_cycles_flatten = set(), [], []
    sub_adj_matrix = full_adjacency_matrix[cc_trace]
    sub_adj_matrix = sub_adj_matrix[:, cc_trace]
    dfs_fc(0, [0], sub_adj_matrix)

    is_cycle_boarder = [False] * len(cc_trace)
    for cycle in cycles:
        num_edges = (sub_adj_matrix[list(cycle)] > 0).sum(-1)
        for i, c in enumerate(cycle):
            if num_edges[i] < 3:
                is_cycle_boarder[cycle[i]] = True

    return is_cycle_boarder


def recursive_trim(cc_trace, cc_score_total, graph_pack, reaction_centers, total_num=None,
                   min_count=2, max_count=25, num_removal=3, depth=0):
    """Recursively prune a root graph into its sub-graphs based on heuristics"""
    if tuple(sorted(cc_trace)) in reaction_centers:
        return
    if len(cc_trace) < min_count:
        return

    node_scores, edge_scores, adj_matrix, full_adj_matrix = graph_pack
    sub_adj_matrix = full_adj_matrix[cc_trace]
    sub_adj_matrix = sub_adj_matrix[:, cc_trace]
    if total_num is None:
        total_num = (sub_adj_matrix > 0).sum().item() + len(cc_trace)

    is_boarder = ((sub_adj_matrix > 0).sum(0) == 1)
    is_cycle_boarder = get_boarder_cycles(cc_trace, full_adj_matrix)

    idx2score = {}
    cc_node_scores = node_scores[cc_trace]
    for idx in range(len(cc_trace)):
        if is_boarder[idx] or is_cycle_boarder[idx]:
            idx2score[idx] = cc_node_scores[idx]

    top_index_by_sorted_node_scores = sorted(list(idx2score.keys()), key=lambda x: idx2score[x])

    if len(cc_trace) > max_count:
        top_index_by_sorted_node_scores = top_index_by_sorted_node_scores[:1]
        depth_update = depth
    elif depth > 2:
        top_index_by_sorted_node_scores = top_index_by_sorted_node_scores[:num_removal]
        depth_update = depth + 1
    else:
        depth_update = depth + 1

    # Normalize score for reaction center
    normalize_factor = len(cc_trace) + np.sum(sub_adj_matrix > 0)
    reaction_centers[tuple(sorted(cc_trace))] = cc_score_total / normalize_factor

    for idx in top_index_by_sorted_node_scores:
        cc_trace_next = cc_trace[:]
        cc_trace_next.pop(idx)
        cc_score_next = cc_score_total
        js = [cc_trace[idx_j] for idx_j in np.argwhere(sub_adj_matrix[idx] > 0)[0]]
        cc_score_next -= np.sum(np.log(edge_scores[js, cc_trace[idx]]))
        cc_score_next -= np.sum(np.log(edge_scores[cc_trace[idx], js]))
        cc_score_next -= np.log(node_scores[cc_trace[idx]])

        recursive_trim(cc_trace_next, cc_score_next, graph_pack, reaction_centers,
                       total_num=total_num, min_count=min_count, max_count=max_count,
                       num_removal=num_removal, depth=depth_update)
    return


def get_subgraphs_by_trim(cc_trace, cc_score_total, graph_pack,
                          min_count=2, max_count=25, num_removal=3, verbose=False):
    """Wrapper for recursive pruning"""
    if num_removal < 0:
        num_removal = len(cc_trace)
    reaction_centers = {}

    recursive_trim(cc_trace, cc_score_total, graph_pack, reaction_centers,
                   min_count=min_count, max_count=max_count, num_removal=num_removal)
    if verbose:
        print('trim:')
        tmp = sorted(reaction_centers.items(), key=lambda x: -x[1])
        for t in tmp:
            print('  {}'.format(t))

    return list(reaction_centers.keys())


def add_mapping(token, map_num=1):
    """Add a given tag (atom mapping) into a SMILES token"""
    if not re.match('.*[a-zA-Z].*', token):
        return token
    if token[0] == '[':
        if re.match('\[.*:[1-9]+\]', token):
            result = token
        else:
            result = token.replace(']', ':{}]'.format(map_num))
    else:
        result = '[{}:{}]'.format(token, map_num)
    return result


def clear_map_number(smi):
    """Clear the atom mapping number of a SMILES sequence"""
    mol = Chem.MolFromSmiles(smi)
    for atom in mol.GetAtoms():
        if atom.HasProp('molAtomMapNumber'):
            atom.ClearProp('molAtomMapNumber')
    return canonical_smiles(Chem.MolToSmiles(mol))


def canonical_smiles(smi):
    """Canonicalize a SMILES without atom mapping"""
    mol = Chem.MolFromSmiles(smi)
    if mol is None:
        return smi
    else:
        canonical_smi = Chem.MolToSmiles(mol)
        # print('>>', canonical_smi)
        if '.' in canonical_smi:
            canonical_smi_list = canonical_smi.split('.')
            canonical_smi_list = sorted(canonical_smi_list, key=lambda x: (len(x), x))
            canonical_smi = '.'.join(canonical_smi_list)
        return canonical_smi


def randomize_smiles_with_am(smi):
    """Randomize a SMILES with atom mapping"""
    mol = Chem.MolFromSmiles(smi)
    random_root = np.random.choice([(atom.GetIdx()) for atom in mol.GetAtoms()])
    return Chem.MolToSmiles(mol, rootedAtAtom=int(random_root))


def canonical_smiles_with_am(smi):
    """Canonicalize a SMILES with atom mapping"""
    atomIdx2am, pivot2atomIdx = {}, {}
    mol = Chem.MolFromSmiles(smi)
    atom_ordering = []
    for atom in mol.GetAtoms():
        if atom.HasProp('molAtomMapNumber'):
            atomIdx2am[atom.GetIdx()] = atom.GetProp('molAtomMapNumber')
            atom.ClearProp('molAtomMapNumber')
        else:
            atomIdx2am[atom.GetIdx()] = '0'
        atom_ordering.append(atom.GetIdx())

    unmapped_smi = Chem.MolFragmentToSmiles(mol, atomsToUse=atom_ordering, canonical=False)
    mol = Chem.MolFromSmiles(unmapped_smi)
    cano_atom_ordering = list(Chem.CanonicalRankAtoms(mol))

    for i, j in enumerate(cano_atom_ordering):
        pivot2atomIdx[j + 1] = i
        mol.GetAtomWithIdx(i).SetIntProp('molAtomMapNumber', j + 1)

    new_tokens = []
    for token in smi_tokenizer(Chem.MolToSmiles(mol)):
        if re.match('.*:([0-9]+)]', token):
            pivot = re.match('.*(:[0-9]+])', token).group(1)
            token = token.replace(pivot, ':{}]'.format(atomIdx2am[pivot2atomIdx[int(pivot[1:-1])]]))
        new_tokens.append(token)

    canonical_smi = ''.join(new_tokens)
    # canonical reactants order
    if '.' in canonical_smi:
        canonical_smi_list = canonical_smi.split('.')
        canonical_smi_list = sorted(canonical_smi_list, key=lambda x: (len(x), x))
        canonical_smi = '.'.join(canonical_smi_list)
    return canonical_smi


def smi_tokenizer(smi):
    """Tokenize a SMILES sequence or reaction"""
    pattern = "(\[[^\]]+]|Bi|Br?|Ge|Te|Mo|K|Ti|Zr|Y|Na|125I|Al|Ce|Cr|Cl?|Ni?|O|S|Pd?|Fe?|I|b|c|Mn|n|o|s|<unk>|>>|Li|p|\(|\)|\.|=|#|-|\+|\\\\|\/|:|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])"
    regex = re.compile(pattern)
    tokens = [token for token in regex.findall(smi)]
    if smi != ''.join(tokens):
        print('ERROR:', smi, ''.join(tokens))
    assert smi == ''.join(tokens)
    return tokens


def remove_am_without_canonical(smi_am, force_canonical=False):
    """Get the canonical SMILES by token modification (smiles arranged by CanonicalRankAtoms)
    :param smi_am: SMILES from `canonical_smiles_with_am`
    :param force_canonical: force the output to be canonical, not recommended since it may break the alignment
    :return:
    """

    def check_special_token(token):
        pattern = "(Mg|Zn|Si|Sn|Se|se|Ge|K|Ti|Pd|Mo|Ce|Ta|As|te|Pb|Ru|Ag|W|Pt|Co|Ca|Xe|11CH3|Rh|Tl|V|131I|Re|13c|siH|La|pH|Y|Zr|Bi|125I|Sb|Te|Ni|Fe|Mn|Cr|Al|Na|Li|Cu|nH[0-9]?|NH[1-9]?\+|\+|-|@|PH[1-9]?)"
        regex = re.compile(pattern)
        return regex.findall(token)

    new_tokens = []
    for token in smi_tokenizer(smi_am):
        # Has atommapping:
        if token[0] == '[' and token[-1] == ']' and re.match('.*:([0-9]+)]', token):
            # print(token)
            token = token.replace(re.match('.*(:[0-9]+)]', token).group(1), '')
            explicitHs = re.match('.*(H[1-9]?).*', token)
            onlyH = re.match('\[[1-9]?H', token)
            if explicitHs and not check_special_token(token) and not onlyH:
                token = token.replace(explicitHs.group(1), '')[1:-1]
            elif not check_special_token(token) and not onlyH:
                token = token[1:-1]
            else:
                token = token

            # print(token)
            # print( )
        new_tokens.append(token)

    canonical_smi = ''.join(new_tokens)
    if force_canonical:
        canonical_smi = canonical_smiles(canonical_smi)
    return canonical_smi


def extract_relative_mapping(cano_prod_am, cano_reacts_am):
    """Extract the reactants relative positional mapping based on SMILES from `canonical_smiles_with_am`
    :param cano_prod_am:
    :param cano_reacts_am:
    :return:
    """
    cano_prod_tokens = smi_tokenizer(cano_prod_am)
    cano_reacts_tokens = smi_tokenizer(cano_reacts_am)

    # Get the exact token mapping for canonical smiles
    prodToken2posIdx = {}
    for i, token in enumerate(cano_prod_tokens):
        if token[0] == '[' and token[-1] == ']' and re.match('.*:([0-9]+)]', token):
            am = int(re.match('.*:([0-9]+)]', token).group(1))
            prodToken2posIdx[am] = i
        else:
            prodToken2posIdx[token] = prodToken2posIdx.get(token, []) + [i]
    position_mapping_list = []
    for i, token in enumerate(cano_reacts_tokens):
        if token[0] == '[' and token[-1] == ']' and re.match('.*:([0-9]+)]', token):
            am = int(re.match('.*:([0-9]+)]', token).group(1))
            prod_posIdx = prodToken2posIdx.get(am, -1)

            if prod_posIdx != -1:
                if (i, prod_posIdx) not in position_mapping_list:
                    position_mapping_list.append((i, prod_posIdx))
                    # print(i, prod_posIdx, cano_reacts_tokens[i], cano_prod_tokens[prod_posIdx])

            # Try to expand with increment
            react_pivot = i + 1
            prod_pivot = prod_posIdx + 1
            while (react_pivot, prod_pivot) not in position_mapping_list and \
                    react_pivot < len(cano_reacts_tokens) and prod_pivot < len(cano_prod_tokens) and \
                    cano_reacts_tokens[react_pivot] == cano_prod_tokens[prod_pivot]:
                # print(react_pivot, prod_pivot, cano_reacts_tokens[react_pivot])
                position_mapping_list.append((react_pivot, prod_pivot))
                react_pivot += 1
                prod_pivot += 1


            # Try to expand with decrement
            react_pivot = i - 1
            prod_pivot = prod_posIdx - 1
            while (react_pivot, prod_pivot) not in position_mapping_list and \
                    react_pivot > -1 and prod_pivot > -1 and \
                    cano_reacts_tokens[react_pivot] == cano_prod_tokens[prod_pivot]:
                position_mapping_list.append((react_pivot, prod_pivot))
                # print(react_pivot, prod_pivot, cano_reacts_tokens[react_pivot])
                react_pivot -= 1
                prod_pivot -= 1


    return position_mapping_list


def get_nonreactive_mask(cano_prod_am, raw_prod, raw_reacts, radius=0):
    """Retrieve the ground truth reaction center by RDChiral"""
    reactants = mols_from_smiles_list(replace_deuterated(raw_reacts).split('.'))
    products = mols_from_smiles_list(replace_deuterated(raw_prod).split('.'))
    changed_atoms, changed_atom_tags, err = get_changed_atoms(reactants, products)
    # print(changed_atom_tags)

    for _ in range(radius):
        mol = Chem.MolFromSmiles(cano_prod_am)
        changed_atom_tags_neighbor = []
        for atom in mol.GetAtoms():
            if atom.GetSmarts().split(':')[1][:-1] in changed_atom_tags:
                for n_atom in atom.GetNeighbors():
                    changed_atom_tags_neighbor.append(n_atom.GetSmarts().split(':')[1][:-1])
        changed_atom_tags = list(set(changed_atom_tags + changed_atom_tags_neighbor))
    # print(changed_atom_tags)

    nonreactive_mask = []
    for i, token in enumerate(smi_tokenizer(cano_prod_am)):
        if token[0] == '[' and token[-1] == ']' and re.match('.*:([0-9]+)]', token):
            am = re.match('.*:([0-9]+)]', token).group(1)
            if am in changed_atom_tags:
                nonreactive_mask.append(False)
                continue
        nonreactive_mask.append(True)

    if sum(nonreactive_mask) == len(nonreactive_mask):  # if the reaction center is not detected
        nonreactive_mask = [False] * len(nonreactive_mask)
    return nonreactive_mask



import math
from time import time
import numpy as np
import pickle
from easydict import EasyDict

import os
from collections import defaultdict

import warnings
warnings.filterwarnings("ignore")

import sys
root = os.path.abspath('.')
if root not in sys.path:
    sys.path.insert(0, root)

from retroformer.utils.smiles_utils import *
from retroformer.utils.translate_utils import translate_batch_stepwise
from retroformer.utils.build_utils import build_model, load_checkpoint, build_batch

from rdkit import Chem
    

def filter_invalid(generations, scores):
    filtered_generations, filtered_scores = [], []
    for generation, score in zip(generations, scores):
        mol = Chem.MolFromSmiles(generation)
        if mol is None:
            continue
        filtered_generations.append(Chem.MolToSmiles(mol))
        filtered_scores.append(score)
    
    return filtered_generations, filtered_scores


def filter_single_atom(generations):
    filtered_generations = []
    for generation in generations:
        filtered_generations.append('.'.join([smi for smi in generation.split('.') if len(smi) > 1]))
    
    return filtered_generations


def translate(args, batch, models, dataset, singlestep=False):
    generations, scores = [], []
    invalid_token_indices = [dataset.tgt_stoi['<RX_{}>'.format(i)] for i in range(1, 11)]
    invalid_token_indices += [dataset.tgt_stoi['<UNK>'], dataset.tgt_stoi['<unk>']]
    
    generation_dict, counter = defaultdict(int), defaultdict(int)
    for model in models:
        pred_tokens, pred_scores, predicts = \
            translate_batch_stepwise(model, batch, invalid_token_indices=invalid_token_indices,
                                    beam_size=args.beam_size,
                                    T=10, alpha_atom=-1, alpha_bond=-1,
                                    beta=0.5, percent_aa=40, percent_ab=40, k=3,
                                    use_template=False,
                                    factor_func=dataset.factor_func,
                                    reconstruct_func=dataset.reconstruct_smi)
        if not singlestep:
            assert len(predicts) == 1

        original_beam_size = pred_tokens.shape[1]
        current_i = 0
        for predict in predicts:
            remain = original_beam_size
            beam_size = math.ceil(original_beam_size / len(predict))

            hypo_i, hypo_scores_i = [], []
            for j, (rc, rc_score) in enumerate(predict):
                pred_token = pred_tokens[current_i + j]

                sub_hypo_candidates, sub_score_candidates = [], []
                for k in range(pred_token.shape[0]):
                    hypo_smiles_k = ''.join(dataset.reconstruct_smi(pred_token[k], src=False))
                    hypo_lens_k = len(smi_tokenizer(hypo_smiles_k))
                    hypo_scores_k = pred_scores[current_i + j][k].cpu().numpy() / hypo_lens_k + rc_score / 10

                    if hypo_smiles_k not in hypo_i:  # only select unique entries
                        sub_hypo_candidates.append(hypo_smiles_k)
                        sub_score_candidates.append(hypo_scores_k)

                ordering = np.argsort(sub_score_candidates)[::-1]
                sub_hypo_candidates = list(np.array(sub_hypo_candidates)[ordering])[:min(beam_size, remain)]
                sub_score_candidates = list(np.array(sub_score_candidates)[ordering])[:min(beam_size, remain)]

                hypo_i += sub_hypo_candidates
                hypo_scores_i += sub_score_candidates

                remain -= beam_size

            current_i += len(predict)
            ordering = np.argsort(hypo_scores_i)[::-1]
            generations.append(np.array(hypo_i)[ordering][:10])
            scores.append(np.array(hypo_scores_i)[ordering][:10])
        
        if singlestep:
            generations = [list(g) for g in generations]
            scores = [list(s) for s in scores]
            for i in range(len(scores)):
                if len(scores[i]) > 0:
                    scores[i] = list(np.exp(np.array(scores[i])) / np.sum(np.exp(np.array(scores[i]))))     # softmax
            return generations, scores

        _generations = list(np.array(hypo_i)[ordering])
        _scores = list(np.array(hypo_scores_i)[ordering])

        _generations, _scores = filter_invalid(_generations, _scores)
        _generations = filter_single_atom(_generations)

        for i, (g, s) in enumerate(zip(_generations, _scores)):
            generation_dict[g] += s
            counter[g] += 1

    for g, cnt in counter.items():
        generation_dict[g] /= cnt
    
    generations = list(generation_dict.keys())
    scores = list(generation_dict.values())
    
    ordering = np.argsort(scores)[::-1][:args.expansion_topk]
    generations = list(np.array(generations)[ordering])
    scores = list(np.array(scores)[ordering])

    if len(scores) > 0:
        scores = list(np.exp(np.array(scores)) / np.sum(np.exp(np.array(scores))))   # softmax

    return {'reactants': generations, 'scores': scores}


def prepare_retroformer(cuda=True, beam_size=10, expansion_topk=10):
    args = EasyDict()
    args.device = 'cuda' if cuda else 'cpu'
    args.beam_size = beam_size
    args.expansion_topk = expansion_topk
    args.vocab = 'retroformer/saved_models/vocab_share.pk'
    # args.checkpoint = 'retroformer/saved_models/clean.pt'
    args.checkpoint = 'retroformer/saved_models/biochem.pt'
    
    with open(args.vocab, 'rb') as f:
        src_itos, tgt_itos = pickle.load(f)

    # ensemble
    args.checkpoint = args.checkpoint.split(',')
    models = [build_model(args, src_itos, tgt_itos) for _ in range(len(args.checkpoint))]
    models = load_checkpoint(args, models)
    return models, args


def run_retroformer(model, args, smi):
    singlestep = True
    if type(smi) == str:
        smi = [smi]
        singlestep = False
    batch, dataset = build_batch(args, smi)
    result = translate(args, batch, model, dataset, singlestep)
    if singlestep:
        return result
    result['templates'] = [None for _ in range(len(result['scores']))]
    result['retrieved'] = [False for _ in range(len(result['scores']))]
    return result


if __name__ == '__main__':
    t_start = time()
    model, args = prepare_retroformer()
    # smi = 'C=CC(C)(C)[C@@]12C[C@H]3c4nc5ccccc5c(=O)n4[C@H](C)C(=O)N3C1N(C(C)=O)c1ccccc12'
    smi = ['C=CC(C)(C)[C@@]12C[C@H]3c4nc5ccccc5c(=O)n4[C@H](C)C(=O)N3C1N(C(C)=O)c1ccccc12', 'N[C@@H](CNC(=O)C(=O)O)C(=O)O']
    print(run_retroformer(model, args, smi))
    print(f'\033[92mTotal {time() - t_start:.2f} sec elapsed\033[0m')



import torch
import torch.nn as nn
from g2s.models.dgcn import DGCNEncoder
from typing import Tuple
from g2s.utils.data_utils import G2SBatch


class GraphFeatEncoder(nn.Module):
    def __init__(self, args, n_atom_feat: int, n_bond_feat: int):
        super().__init__()
        self.args = args

        self.n_atom_feat = n_atom_feat
        self.n_bond_feat = n_bond_feat

        self.mpn = DGCNEncoder(
            args,
            input_size=n_atom_feat + n_bond_feat,
            node_fdim=n_atom_feat
        )

    def forward(self, reaction_batch: G2SBatch) -> Tuple[torch.Tensor, None]:
        """
        Forward pass of the graph encoder. First the feature vectors are extracted,
        and then encoded. This has been modified to pass data via the G2SBatch datatype
        """
        fnode = reaction_batch.fnode
        fmess = reaction_batch.fmess
        agraph = reaction_batch.agraph
        bgraph = reaction_batch.bgraph

        # embed graph, note that for directed graph, fess[any, 0:2] = u, v
        hnode = fnode.clone()
        fmess1 = hnode.index_select(index=fmess[:, 0].long(), dim=0)
        fmess2 = fmess[:, 2:].clone()
        hmess = torch.cat([fmess1, fmess2], dim=-1)         # hmess = x = [x_u; x_uv]

        # encode
        hatom, _ = self.mpn(hnode, hmess, agraph, bgraph, mask=None)
        hmol = None

        return hatom, hmol



import logging


def setup_logger(fname=None, silent=False):
    if fname is None:
        logging.basicConfig(
            level=logging.INFO if not silent else logging.CRITICAL,
            format='%(name)-12s: %(levelname)-8s %(message)s',
            datefmt='%m-%d %H:%M',
            filemode='w'
        )
    else:
        logging.basicConfig(
            level=logging.INFO if not silent else logging.CRITICAL,
            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
            datefmt='%m-%d %H:%M',
            filename=fname,
            filemode='w'
        )
        console = logging.StreamHandler()
        console.setLevel(logging.INFO)
        formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
        console.setFormatter(formatter)
        logging.getLogger('').addHandler(console)



import numpy as np

class SynRoute:
    def __init__(self, target_mol, succ_value, search_status):
        self.target_mol = target_mol
        self.mols = [target_mol]
        self.values = [None]
        self.templates = [None]
        self.parents = [-1]
        self.children = [None]
        self.optimal = False
        self.costs = {}

        self.succ_value = succ_value
        self.total_cost = 0
        self.length = 0
        self.search_status = search_status
        if self.succ_value <= self.search_status:
            self.optimal = True

    def _add_mol(self, mol, parent_id):
        self.mols.append(mol)
        self.values.append(None)
        self.templates.append(None)
        self.parents.append(parent_id)
        self.children.append(None)

        self.children[parent_id].append(len(self.mols)-1)

    def set_value(self, mol, value):
        assert mol in self.mols

        mol_id = self.mols.index(mol)
        self.values[mol_id] = value

    def add_reaction(self, mol, value, template, reactants, cost):
        assert mol in self.mols

        self.total_cost += cost
        self.length += 1

        parent_id = self.mols.index(mol)
        self.values[parent_id] = value
        self.templates[parent_id] = template
        self.children[parent_id] = []
        self.costs[parent_id] = cost

        for reactant in reactants:
            self._add_mol(reactant, parent_id)

    def serialize_reaction(self, idx):
        s = self.mols[idx]
        if self.children[idx] is None:
            return s
        s += '>%.4f>' % np.exp(-self.costs[idx])
        s += self.mols[self.children[idx][0]]
        for i in range(1, len(self.children[idx])):
            s += '.'
            s += self.mols[self.children[idx][i]]

        return s

    def serialize(self):
        s = self.serialize_reaction(0)
        for i in range(1, len(self.mols)):
            if self.children[i] is not None:
                s += '|'
                s += self.serialize_reaction(i)

        return s



from retro_star.api import RSPlanner
import sys


def worker(q, thread_id, args, start_idx, end_idx):
    msg = run_multistep(thread_id, args, start_idx, end_idx)
    q.put(msg)


def listener(q, save_name):
    with open(save_name, 'a') as f:
        while True:
            m = q.get()
            if m == '#done#':
                break
            f.write(str(m))
            f.flush()


def run_multistep(thread_id, args, mol_ids, products):
    print(f'[Thread {thread_id}] start')
    planner = RSPlanner(
        cuda=args.device=='cuda',
        iterations=args.iterations,
        expansion_topk=args.exp_topk,
        route_topk=args.route_topk,
        beam_size=args.beam_size,
        model_type=args.model_type,
        retrieval=args.retrieval=='true',
        retrieval_db=args.retrieval_db,
        path_retrieval=args.path_retrieval=='true',
        path_retrieval_db=args.path_retrieval_db,
        starting_molecules=args.blocks
    )
    msg = ''
    done = 0
    for mol_id, product in zip(mol_ids, products):
        try:
            result = planner.plan(product)
        except KeyboardInterrupt:
            print('KeyboardInterrupt')
            sys.exit()
        except:
            result = 'Error'

        if result is None:
            msg += f'{mol_id} None\n'
        elif result == 'Error':
            msg += f'{mol_id} Error\n'
        else:
            for i, route in enumerate(result):
                msg += f'{mol_id} {i} {route}\n'
        done += 1
        print(f'[Thread {thread_id}] {done}/{len(products)}')
    print(f'[Thread {thread_id}] Finished')
    return msg



import torch
import torch.nn as nn
from g2s.models.model_utils import index_select_ND
from typing import Tuple


class DGCNGRU(nn.Module):
    """GRU Message Passing layer."""
    def __init__(self, args, input_size: int, h_size: int, depth: int):
        super().__init__()
        self.args = args

        self.input_size = input_size
        self.h_size = h_size
        self.depth = depth

        self._build_layer_components()

    def _build_layer_components(self) -> None:
        """Build layer components."""
        self.W_z = nn.Linear(self.input_size + self.h_size, self.h_size)
        self.W_r = nn.Linear(self.input_size, self.h_size, bias=False)
        self.U_r = nn.Linear(self.h_size, self.h_size)
        self.W_h = nn.Linear(self.input_size + self.h_size, self.h_size)

    def GRU(self, x: torch.Tensor, h_nei: torch.Tensor) -> torch.Tensor:
        """Implements the GRU gating equations.

        Parameters
        ----------
            x: torch.Tensor, input tensor
            h_nei: torch.Tensor, hidden states of the neighbors
        """
        sum_h = h_nei.sum(dim=1)                        # (9)
        z_input = torch.cat([x, sum_h], dim=1)          # x = [x_u; x_uv]
        z = torch.sigmoid(self.W_z(z_input))            # (10)

        r_1 = self.W_r(x).view(-1, 1, self.h_size)
        r_2 = self.U_r(h_nei)
        r = torch.sigmoid(r_1 + r_2)                    # (11) r_ku = f_r(x; m_ku) = W_r(x) + U_r(m_ku)

        gated_h = r * h_nei
        sum_gated_h = gated_h.sum(dim=1)                # (12)
        h_input = torch.cat([x, sum_gated_h], dim=1)
        pre_h = torch.tanh(self.W_h(h_input))           # (13)
        new_h = (1.0 - z) * sum_h + z * pre_h           # (14)

        return new_h

    def forward(self, fmess: torch.Tensor, bgraph: torch.Tensor) -> torch.Tensor:
        """Forward pass of the RNN

        Parameters
        ----------
            fmess: torch.Tensor, contains the initial features passed as messages
            bgraph: torch.Tensor, bond graph tensor. Contains who passes messages to whom.
        """
        h = torch.zeros(fmess.size()[0], self.h_size, device=fmess.device)
        mask = torch.ones(h.size()[0], 1, device=h.device)
        mask[0, 0] = 0      # first message is padding

        for i in range(self.depth):
            h_nei = index_select_ND(h, 0, bgraph)
            h = self.GRU(fmess, h_nei)
            h = h * mask
        return h


class DGCNEncoder(nn.Module):
    """MessagePassing Network based encoder. Messages are updated using an RNN
    and the final message is used to update atom embeddings."""
    def __init__(self, args, input_size: int, node_fdim: int):
        super().__init__()
        self.args = args

        self.h_size = args.encoder_hidden_size
        self.depth = args.encoder_num_layers
        self.input_size = input_size
        self.node_fdim = node_fdim

        self._build_layers()

    def _build_layers(self) -> None:
        """Build layers associated with the MPNEncoder."""
        self.W_o = nn.Sequential(nn.Linear(self.node_fdim + self.h_size, self.h_size), nn.GELU())
        self.rnn = DGCNGRU(self.args, self.input_size, self.h_size, self.depth)

    def forward(self, fnode: torch.Tensor, fmess: torch.Tensor,
                agraph: torch.Tensor, bgraph: torch.Tensor, mask: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """Forward pass of the MPNEncoder.

        Parameters
        ----------
            fnode: torch.Tensor, node feature tensor
            fmess: torch.Tensor, message features
            agraph: torch.Tensor, neighborhood of an atom
            bgraph: torch.Tensor, neighborhood of a bond,
                except the directed bond from the destination node to the source node
            mask: torch.Tensor, masks on nodes
        """
        h = self.rnn(fmess, bgraph)
        nei_message = index_select_ND(h, 0, agraph)
        nei_message = nei_message.sum(dim=1)
        node_hiddens = torch.cat([fnode, nei_message], dim=1)
        node_hiddens = self.W_o(node_hiddens)

        if mask is None:
            mask = torch.ones(node_hiddens.size(0), 1, device=fnode.device)
            mask[0, 0] = 0      # first node is padding

        return node_hiddens * mask, h



import argparse
from rdkit import Chem

def rm_stereo(smi):
    try:
        return Chem.MolToSmiles(Chem.MolFromSmiles(smi),isomericSmiles=False)
    except:
        import pdb; pdb.set_trace()
        print(smi)
        return smi

def dict_update(tdict, tkey, tval):
    if tkey in tdict:
        pos = len(tdict[tkey])
        tdict[tkey].update({pos: tval})
    else:
        tdict[tkey] = {0: tval}


def get_pred_dict(pth, mol2class, product_class='all'):
    mol_dict= {}
    with open(pth, 'r') as f:
        lines = [i.strip('\n') for i in f.readlines()]
    for line in lines:
        path = line.split(' ')[-1]
        if path in {'None', 'Error'}:
            continue
        product = path.split('>')[0]
        if product_class == 'all' or mol2class[product] == product_class:
            dict_update(mol_dict, product, path)
    return mol_dict


def get_ground_truth_dict(pth):
    '''
    :param pth: the path of ground truth file(.txt)
    :return: a dict, format: {
                    smiles1: {0: ground truth path1; 1: ground truth path2; ...}
                    smiles2: {0: ground truth path1; 1: ground truth path2; ...}
                }
    '''
    ground_truth_dict = {}
    with open(pth, 'r') as f:
        l = [i.strip('\n') for i in f.readlines()]
    
    for line in l:
        ele = line.split('\t')
        ele = [i for i in ele if len(i) != 0]
        target = ele[3]
        dict_update(ground_truth_dict, target, '|'.join(ele[3:]))
    return ground_truth_dict


def get_lst_from_pred_dict(s):  
    '''
    :param s: 'smiles1>score1>smiles2|smiles2>score2>smiles3|smiles3>score3>smiles4...'
    :return: [smiles1, smiles2, smiles3, ...]
    '''
    lst = []
    rea = s.split('|')
    for i in range(len(rea)):
        line = rea[i].split('>')
        if i == 0:
            lst.append(line[0])
        lst.append(line[-1])
        #     lst.append(rm_stereo(line[0]))
        # lst.append(rm_stereo(line[-1]))
    return lst


def get_lst_from_gt_dict(s):
    '''
    :param s: 'smiles1|smiles2|smiles3...'
    :return: [smiles1, smiles2, smiles3, ...]
    '''
    l = s.split('|')
    mols = []
    for i in l:
        try:
            mols.append(i)
        except:
            print(f"wrong reaction: {i}")
    return mols


def blockHit(ori_path, pred_path):
    ori_bb = ori_path[-1]
    pred_bb = pred_path[-1]
    return ori_bb == pred_bb


def pathHit(ori_path, pred_path):
    ori_set = set(ori_path)
    pred_set = set(pred_path)
    intersect = len(ori_set & pred_set)
    num = 0
    # if len(ori_path) == len(pred_path) and intersect == len(ori_path):
    if intersect == len(ori_path):
        num = 1
    return num


class MultiPredRes:
    def __init__(self, pred_p, testSetLength):
        self.successPath = len(pred_p)
        self.blockHit = 0
        self.pathHit = 0
        self.testSetLen = testSetLength
        self.result_dict = {}

    def resultShow(self):
        s =  f"Success rate:\t\t\t{self.successPath / self.testSetLen*100:.4f}%\n"
        s += f"Hit rate of building blocks:\t{self.blockHit / self.testSetLen*100:.4f}%\n"
        s += f"Hit rate of pathways:\t\t{self.pathHit / self.testSetLen*100:.4f}%"
        print(s)


def multiVal(pred_path, product_class='all'):
    with open('data/test.txt') as f:
        mol2class = f.readlines()
        
    # mol2class = [l.strip().split('\t') for l in mol2class]
    mol2class = [l.strip().split() for l in mol2class]
    # mol2class = {l[2]: l[1] for l in mol2class}
    mol2class = {l[-1]: l[1] for l in mol2class}

    products = set()
    for product, _class in mol2class.items():
        if product_class == 'all' or _class == product_class:
            products.add(product)
            
    testSetNum = len(products)
    print(f'Number of test molecules:\t{testSetNum}')

    pred_dict = get_pred_dict(pred_path, mol2class, product_class)
    ground_truth_dict = get_ground_truth_dict('data/test_gt.txt')

    mRes = MultiPredRes(pred_dict, testSetNum)

    for smiles, v_dict in pred_dict.items():
        if v_dict is None:
            continue
        try:
            gt_dict = ground_truth_dict[smiles]
        except:
            # import pdb; pdb.set_trace()
            print("fail" + smiles)
            print('Maybe the streo rm form product was used in model')
        building_block_same = False
        path_same = False
        for routes_score in v_dict.values():
            for item in gt_dict.values():
                gt_lst = get_lst_from_gt_dict(item)
                pred_lst = get_lst_from_pred_dict(routes_score)

                bb_same = blockHit(gt_lst, pred_lst)
                if bb_same:
                    building_block_same = True

                num = pathHit(gt_lst, pred_lst)
                if num:
                    path_same = True

        if building_block_same:
            mRes.blockHit += 1
        if path_same:
            mRes.pathHit += 1
    mRes.resultShow()
    # import pdb; pdb.set_trace()

parser = argparse.ArgumentParser()
parser.add_argument('save_file', type=str)
args = parser.parse_args()

multiVal(args.save_file)



from tqdm import tqdm
from rdkit import Chem
import pandas as pd
import os, sys

def gen_pathways(path_path, compound_path):
    
    with open(compound_path) as f:
        compounds =  f.readlines()
        compounds = {i.rstrip().split()[0]:i.rstrip().split()[1] for i in compounds[1:]}
    with open(path_path) as f:
        path_lines = f.readlines()
        path_lines = [i.rstrip().split(',') for i in path_lines[1:]]
    # print(compounds)
    starting = Chem.MolToSmiles(Chem.MolFromSmiles(compounds[path_lines[0][4].split('.')[-1].replace('"','')]))
    pathways = []
    pathway, id = [], '0'
    for line in path_lines:
        if id != line[0]:
            pathways.append(pathway)
            id = line[0]
            pathway = [starting]
        try:
            pathway.append(Chem.MolToSmiles(Chem.MolFromSmiles(compounds[line[-2].split('.')[-1].replace('"','')])))
        except:
            print(line[3])
    pathways = pathways[1:]
        
    return pathways

def dict_update(tdict, tkey, tval):
    if tkey in tdict:
        pos = len(tdict[tkey])
        tdict[tkey].update({pos: tval})
    else:
        tdict[tkey] = {0: tval}

def get_lst_from_pred_dict(s):  
    '''
    :param s: 'smiles1>score1>smiles2|smiles2>score2>smiles3|smiles3>score3>smiles4...'
    :return: [smiles1, smiles2, smiles3, ...]
    '''
    lst = []
    rea = s.split('|')
    for i in range(len(rea)):
        line = rea[i].split('>')
        if i == 0:
            lst.append(line[0])
        lst.append(line[-1])
    return lst

def get_lst_from_gt_dict(s):
    '''
    :param s: 'smiles1|smiles2|smiles3...'
    :return: [smiles1, smiles2, smiles3, ...]
    '''
    l = s.split('|')
    mols = []
    for i in l:
        try:
            mols.append(i)
        except:
            print(f"wrong reaction: {i}")
    return mols

def get_ground_truth_dict(pth):
    '''
    :param pth: the path of ground truth file(.txt)
    :return: a dict, format: {
                    smiles1: {0: ground truth path1; 1: ground truth path2; ...}
                    smiles2: {0: ground truth path1; 1: ground truth path2; ...}
                }
    '''
    ground_truth_dict = {}
    with open(pth, 'r') as f:
        l = [i.strip('\n') for i in f.readlines()]
    
    for line in l:
        ele = line.split('\t')
        ele = [i for i in ele if len(i) != 0]
        target = ele[3]
        dict_update(ground_truth_dict, target, '|'.join(ele[3:]))
    return ground_truth_dict

def blockHit(ori_path, pred_path):
    ori_bb = ori_path[-1]
    pred_bb = pred_path[-1]
    return ori_bb == pred_bb


def pathHit(ori_path, pred_path):
    ori_set = set(ori_path)
    pred_set = set(pred_path)
    intersect = len(ori_set & pred_set)
    num = 0
    if len(ori_path) == len(pred_path) and intersect == len(ori_path):
        num = 1
    return num

class MultiPredRes:
    def __init__(self, testSetLength):
        self.successPath = 0
        self.blockHit = 0
        self.pathHit = 0
        self.testSetLen = testSetLength
        self.result_dict = {}

    def resultShow(self):
        s =  f"Success rate:\t\t\t{self.successPath / self.testSetLen*100:.4f}%\n"
        s += f"Hit rate of building blocks:\t{self.blockHit / self.testSetLen*100:.4f}%\n"
        s += f"Hit rate of pathways:\t\t{self.pathHit / self.testSetLen*100:.4f}%"
        print(s)

if __name__ == '__main__':
    
    root = "/home/taein/Retrosynthesis/rp2paths/examples/paths"
    targets = os.listdir(root)
    ground_truth_dict = get_ground_truth_dict('/home/taein/READRetro/data/test_gt.txt')

    mRes = MultiPredRes(len(targets))
    
    for target in tqdm(targets):
        tmp_results = os.listdir(f"{root}/{target}")
        if 'out_paths.csv' not in tmp_results:
            continue
        mRes.successPath += 1
        pathways = gen_pathways(f'{root}/{target}/out_paths.csv',f'{root}/{target}/compounds.txt')
        smiles = pathways[0][0]
        try:
            gt_dict = ground_truth_dict[smiles]
        except:
            import pdb; pdb.set_trace()
        building_block_same = False
        path_same = False
        
        for path in pathways:
            for item in gt_dict.values():
                gt_lst = get_lst_from_gt_dict(item)

                bb_same = blockHit(gt_lst, path)
                if bb_same:
                    mRes.blockHit += 1

                num = pathHit(gt_lst, path)
                if num:
                    mRes.pathHit += 1
                    
    mRes.resultShow()


from collections import defaultdict
from utils.ensemble import run_ensemble
import pandas as pd
import numpy as np
from rdkit import Chem

def neutralize_atoms(smi):
    try:
        mol = Chem.MolFromSmiles(smi)
        pattern = Chem.MolFromSmarts("[+1!h0!$([*]~[-1,-2,-3,-4]),-1!$([*]~[+1,+2,+3,+4])]")
        at_matches = mol.GetSubstructMatches(pattern)
        at_matches_list = [y[0] for y in at_matches]
        if len(at_matches_list) > 0:
            for at_idx in at_matches_list:
                atom = mol.GetAtomWithIdx(at_idx)
                chg = atom.GetFormalCharge()
                hcount = atom.GetTotalNumHs()
                atom.SetFormalCharge(0)
                atom.SetNumExplicitHs(hcount - chg)
                atom.UpdatePropertyCache()
        return Chem.MolToSmiles(mol,isomericSmiles=False)
    except:
        return smi

def kegg_search(smi,kegg_df):
    extract = kegg_df[kegg_df['SMILES'] == smi]
    if not len(extract):
        return None, None
    else:
        id = extract["ID"].values[0]
        return id, f"www.kegg.jp/entry/{id}"

def common_find(target_key, dict1, dict2):
    common_indices_list1 = [i for i, x in enumerate(dict1[target_key]) if x in dict2[target_key]]
    common_indices_list2 = [i for i, x in enumerate(dict2[target_key]) if x in dict1[target_key]]
    
    common_dict1, common_dict2 = {}, {}
    new_dict1, new_dict2 = {}, {}
    
    for key in dict1:
        values = dict1[key]
        new_dict1[key] = [value for i, value in enumerate(values) if i not in common_indices_list1]
        common_dict1[key] = [value for i, value in enumerate(values) if i in common_indices_list1]

    for key in dict2:
        values = dict2[key]
        new_dict2[key] = [value for i, value in enumerate(values) if i not in common_indices_list2]
        common_dict2[key] = [value for i, value in enumerate(values) if i in common_indices_list2]
    merged_dict = {}

    for key in dict1.keys():
        merged_dict[key] = new_dict1[key] + new_dict2[key]
    return common_dict1, common_dict2, merged_dict

class Retriever:
    def __init__(self, data):
        with open(data) as f:
            reactions = f.readlines()

        self.graph = defaultdict(list)
        for reaction in reactions:
            reactant, product = reaction.strip().split('>>')
            self.graph[product].append(reactant)
    
    def retrieve(self, product):
        if product in self.graph:
            return self.graph[product]

class pathRetriever:
    def __init__(self, kegg_mol_data, pathway_data, token):
        kegg_mol = pd.read_csv(kegg_mol_data)
        pathways = pd.read_pickle(pathway_data)
        compounds = set([i for sublist in pathways["Compounds"] for i in sublist])
        self.compounds = kegg_mol[kegg_mol.apply(lambda row: row['ID'] in compounds, axis=1)]
        self.special_token = token
        
    def retrieve(self, product):
        id, _ = kegg_search(neutralize_atoms(product),self.compounds)
        return [self.special_token] if id != None else []    # if some compounds do not contain in pathways although in compounds, change the condition 



def run_path_retriever(product, path_retriever, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                  vocab, vocab_tokens, device, expansion_topk):
    isPathknown = path_retriever.retrieve(product)
    if isPathknown:
        res_dict = {'reactants': isPathknown}
        res_dict['scores'] = [1. for _ in res_dict['reactants']]
        res_dict['retrieved'] = [True for _ in res_dict['reactants']]
        res_dict['templates'] = [None for _ in res_dict['scores']]
    else:
        res_dict = run_ensemble(product, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                vocab, vocab_tokens, device, expansion_topk)
    return res_dict

def run_retriever(product, retriever, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                  vocab, vocab_tokens, device, expansion_topk):
    reactant_list = retriever.retrieve(product)
    if reactant_list is not None:
        res_dict = {'reactants': reactant_list}
        res_dict['scores'] = [1. for _ in res_dict['reactants']]
        res_dict['retrieved'] = [True for _ in res_dict['reactants']]
        _res_dict = run_ensemble(product, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                 vocab, vocab_tokens, device, expansion_topk)

        res_dict['reactants'].extend(_res_dict['reactants'])
        res_dict['scores'].extend(_res_dict['scores'])
        res_dict['retrieved'].extend(_res_dict['retrieved']) 
        res_dict['templates'] = [None for _ in res_dict['scores']]
        
    else:
        res_dict = run_ensemble(product, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                vocab, vocab_tokens, device, expansion_topk)
    return res_dict

def run_both_retriever(product, path_retriever, retriever, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                  vocab, vocab_tokens, device, expansion_topk):
    isPathknown = path_retriever.retrieve(product)
    if isPathknown:
        res_dict = {'reactants': isPathknown}
        res_dict['scores'] = [1. for _ in res_dict['reactants']]
        res_dict['retrieved'] = [True for _ in res_dict['reactants']]
        res_dict['templates'] = [None for _ in res_dict['scores']]
        return res_dict
        
    reactant_list = retriever.retrieve(product)
    if reactant_list is not None:
        res_dict = {'reactants': reactant_list}
        res_dict['scores'] = [1. for _ in res_dict['reactants']]
        res_dict['retrieved'] = [True for _ in res_dict['reactants']]
        _res_dict = run_ensemble(product, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                 vocab, vocab_tokens, device, expansion_topk)
        
        res_dict['reactants'].extend(_res_dict['reactants'])
        res_dict['scores'].extend(_res_dict['scores'])
        res_dict['retrieved'].extend(_res_dict['retrieved'])
        res_dict['templates'] = [None for _ in res_dict['scores']]
    else:
        res_dict = run_ensemble(product, model_type, model_retroformer, model_g2s, args_retroformer, args_g2s,
                                vocab, vocab_tokens, device, expansion_topk)
    return res_dict

def run_retriever_only(product, path_retriever, retriever):
    isPathknown = path_retriever.retrieve(product)
    if isPathknown:
        res_dict = {'reactants': isPathknown}
        res_dict['scores'] = [1. for _ in res_dict['reactants']]
        res_dict['retrieved'] = [True for _ in res_dict['reactants']]
        res_dict['templates'] = [None for _ in res_dict['scores']]
        return res_dict
    
    reactant_list = retriever.retrieve(product)
    if reactant_list is not None:
        res_dict = {'reactants': reactant_list}
        res_dict['scores'] = [1. for _ in res_dict['reactants']]
        res_dict['retrieved'] = [True for _ in res_dict['reactants']]
        res_dict['templates'] = [None for _ in res_dict['scores']]
    else:
        return {'reactants':[],'scores':[],'retrieved':[],'templates':[]}
    return res_dict



import logging
import os.path

import networkx as nx
import numpy as np
import re
import sys
import time
import torch
from rdkit import Chem
from torch.utils.data import Dataset
from typing import Dict, List, Tuple
from g2s.utils.chem_utils import ATOM_FDIM, BOND_FDIM, get_atom_features_sparse, get_bond_features
from g2s.utils.rxn_graphs import RxnGraph


def tokenize_smiles(smi: str) -> str:
    pattern = r"(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\|\/|:|~|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])"
    regex = re.compile(pattern)
    tokens = [token for token in regex.findall(smi)]
    assert smi == "".join(tokens), f"Tokenization mismatch. smi: {smi}, tokens: {tokens}"

    return " ".join(tokens)


def canonicalize_smiles(smiles, remove_atom_number=False, trim=True, suppress_warning=False):
    cano_smiles = ""

    mol = Chem.MolFromSmiles(smiles)

    if mol is None:
        cano_smiles = ""

    else:
        if trim and mol.GetNumHeavyAtoms() < 2:
            if not suppress_warning:
                logging.info(f"Problematic smiles: {smiles}, setting it to 'CC'")
            cano_smiles = "CC"          # TODO: hardcode to ignore
        else:
            if remove_atom_number:
                [a.ClearProp('molAtomMapNumber') for a in mol.GetAtoms()]
            cano_smiles = Chem.MolToSmiles(mol, isomericSmiles=True)

    return cano_smiles


def len2idx(lens) -> np.ndarray:
    # end_indices = np.cumsum(np.concatenate(lens, axis=0))
    end_indices = np.cumsum(lens)
    start_indices = np.concatenate([[0], end_indices[:-1]], axis=0)
    indices = np.stack([start_indices, end_indices], axis=1)

    return indices


class G2SBatch:
    def __init__(self,
                 fnode: torch.Tensor,
                 fmess: torch.Tensor,
                 agraph: torch.Tensor,
                 bgraph: torch.Tensor,
                 atom_scope: List,
                 bond_scope: List,
                 tgt_token_ids: torch.Tensor,
                 tgt_lengths: torch.Tensor,
                 distances: torch.Tensor = None):
        self.fnode = fnode
        self.fmess = fmess
        self.agraph = agraph
        self.bgraph = bgraph
        self.atom_scope = atom_scope
        self.bond_scope = bond_scope
        self.tgt_token_ids = tgt_token_ids
        self.tgt_lengths = tgt_lengths
        self.distances = distances

        self.size = len(tgt_lengths)

    def to(self, device):
        self.fnode = self.fnode.to(device)
        self.fmess = self.fmess.to(device)
        self.agraph = self.agraph.to(device)
        self.bgraph = self.bgraph.to(device)
        self.tgt_token_ids = self.tgt_token_ids.to(device)
        self.tgt_lengths = self.tgt_lengths.to(device)

        if self.distances is not None:
            self.distances = self.distances.to(device)

    def pin_memory(self):
        self.fnode = self.fnode.pin_memory()
        self.fmess = self.fmess.pin_memory()
        self.agraph = self.agraph.pin_memory()
        self.bgraph = self.bgraph.pin_memory()
        self.tgt_token_ids = self.tgt_token_ids.pin_memory()
        self.tgt_lengths = self.tgt_lengths.pin_memory()

        if self.distances is not None:
            self.distances = self.distances.pin_memory()

        return self

    def log_tensor_shape(self):
        logging.info(f"fnode: {self.fnode.shape}, "
                     f"fmess: {self.fmess.shape}, "
                     f"tgt_token_ids: {self.tgt_token_ids.shape}, "
                     f"tgt_lengths: {self.tgt_lengths}")


class G2SDataset(Dataset):
    def __init__(self, args, file: str):
        self.args = args

        self.a_scopes = []
        self.b_scopes = []
        self.a_features = []
        self.b_features = []
        self.a_graphs = []
        self.b_graphs = []
        self.a_scopes_lens = []
        self.b_scopes_lens = []
        self.a_features_lens = []
        self.b_features_lens = []

        self.src_token_ids = []             # loaded but not batched
        self.src_lens = []
        self.tgt_token_ids = []
        self.tgt_lens = []

        self.data_indices = []
        self.batch_sizes = []
        self.batch_starts = []
        self.batch_ends = []

        self.vocab = load_vocab(args.vocab_file)
        self.vocab_tokens = [k for k, v in sorted(self.vocab.items(), key=lambda tup: tup[1])]

        logging.info(f"Loading preprocessed features from {file}")
        feat = np.load(file)
        for attr in ["a_scopes", "b_scopes", "a_features", "b_features", "a_graphs", "b_graphs",
                     "a_scopes_lens", "b_scopes_lens", "a_features_lens", "b_features_lens",
                     "src_token_ids", "src_lens", "tgt_token_ids", "tgt_lens"]:
            setattr(self, attr, feat[attr])

        # mask out chiral tag (as UNSPECIFIED)
        self.a_features[:, 6] = 2

        assert len(self.a_scopes_lens) == len(self.b_scopes_lens) == \
               len(self.a_features_lens) == len(self.b_features_lens) == \
               len(self.src_token_ids) == len(self.src_lens) == \
               len(self.tgt_token_ids) == len(self.tgt_lens), \
               f"Lengths of source and target mismatch!"

        self.a_scopes_indices = len2idx(self.a_scopes_lens)
        self.b_scopes_indices = len2idx(self.b_scopes_lens)
        self.a_features_indices = len2idx(self.a_features_lens)
        self.b_features_indices = len2idx(self.b_features_lens)

        del self.a_scopes_lens, self.b_scopes_lens, self.a_features_lens, self.b_features_lens

        self.data_size = len(self.src_token_ids)
        self.data_indices = np.arange(self.data_size)

    def sort(self):
        if self.args.verbose:
            start = time.time()

            logging.info(f"Calling G2SDataset.sort()")
            sys.stdout.flush()
            self.data_indices = np.argsort(self.src_lens)

            logging.info(f"Done, time: {time.time() - start: .2f} s")
            sys.stdout.flush()

        else:
            self.data_indices = np.argsort(self.src_lens)

    def shuffle_in_bucket(self, bucket_size: int):
        if self.args.verbose:
            start = time.time()

            logging.info(f"Calling G2SDataset.shuffle_in_bucket()")
            sys.stdout.flush()

            for i in range(0, self.data_size, bucket_size):
                np.random.shuffle(self.data_indices[i:i+bucket_size])

            logging.info(f"Done, time: {time.time() - start: .2f} s")
            sys.stdout.flush()

        else:
            for i in range(0, self.data_size, bucket_size):
                np.random.shuffle(self.data_indices[i:i + bucket_size])

    def batch(self, batch_type: str, batch_size: int):
        start = time.time()

        logging.info(f"Calling G2SDataset.batch()")
        sys.stdout.flush()

        self.batch_sizes = []

        if batch_type == "samples":
            raise NotImplementedError

        elif batch_type == "atoms":
            raise NotImplementedError

        elif batch_type.startswith("tokens"):
            sample_size = 0
            max_batch_src_len = 0
            max_batch_tgt_len = 0

            for data_idx in self.data_indices:
                src_len = self.src_lens[data_idx]
                tgt_len = self.tgt_lens[data_idx]

                max_batch_src_len = max(src_len, max_batch_src_len)
                max_batch_tgt_len = max(tgt_len, max_batch_tgt_len)
                while self.args.enable_amp and not max_batch_src_len % 8 == 0:  # for amp
                    max_batch_src_len += 1
                while self.args.enable_amp and not max_batch_tgt_len % 8 == 0:  # for amp
                    max_batch_tgt_len += 1

                if batch_type == "tokens" and \
                        max_batch_src_len * (sample_size + 1) <= batch_size:
                    sample_size += 1
                elif batch_type == "tokens_sum" and \
                        (max_batch_src_len + max_batch_tgt_len) * (sample_size + 1) <= batch_size:
                    sample_size += 1
                elif self.args.enable_amp and not sample_size % 8 == 0:
                    sample_size += 1
                else:
                    self.batch_sizes.append(sample_size)

                    sample_size = 1
                    max_batch_src_len = src_len
                    max_batch_tgt_len = tgt_len
                    while self.args.enable_amp and not max_batch_src_len % 8 == 0:  # for amp
                        max_batch_src_len += 1
                    while self.args.enable_amp and not max_batch_tgt_len % 8 == 0:  # for amp
                        max_batch_tgt_len += 1

            self.batch_sizes.append(sample_size)
            self.batch_sizes = np.array(self.batch_sizes)
            assert np.sum(self.batch_sizes) == self.data_size, \
                f"Size mismatch! Data size: {self.data_size}, sum batch sizes: {np.sum(self.batch_sizes)}"

            self.batch_ends = np.cumsum(self.batch_sizes)
            self.batch_starts = np.concatenate([[0], self.batch_ends[:-1]])

        else:
            raise ValueError(f"batch_type {batch_type} not supported!")

        logging.info(f"Done, time: {time.time() - start: .2f} s, total batches: {self.__len__()}")
        sys.stdout.flush()

    def __getitem__(self, index: int) -> G2SBatch:
        batch_index = index
        batch_start = self.batch_starts[batch_index]
        batch_end = self.batch_ends[batch_index]

        data_indices = self.data_indices[batch_start:batch_end]

        graph_features = []
        a_lengths = []
        for data_index in data_indices:
            start, end = self.a_scopes_indices[data_index]
            a_scope = self.a_scopes[start:end]
            a_length = a_scope[-1][0] + a_scope[-1][1] - a_scope[0][0]

            start, end = self.b_scopes_indices[data_index]
            b_scope = self.b_scopes[start:end]

            start, end = self.a_features_indices[data_index]
            a_feature = self.a_features[start:end]
            a_graph = self.a_graphs[start:end]

            start, end = self.b_features_indices[data_index]
            b_feature = self.b_features[start:end]
            b_graph = self.b_graphs[start:end]

            graph_feature = (a_scope, b_scope, a_feature, b_feature, a_graph, b_graph)
            graph_features.append(graph_feature)
            a_lengths.append(a_length)

        fnode, fmess, agraph, bgraph, atom_scope, bond_scope = collate_graph_features(graph_features)

        tgt_token_ids = self.tgt_token_ids[data_indices]
        tgt_lengths = self.tgt_lens[data_indices]

        tgt_token_ids = tgt_token_ids[:, :max(tgt_lengths)]

        tgt_token_ids = torch.as_tensor(tgt_token_ids, dtype=torch.long)
        tgt_lengths = torch.tensor(tgt_lengths, dtype=torch.long)

        distances = collate_graph_distances(graph_features, a_lengths)

        g2s_batch = G2SBatch(
            fnode=fnode,
            fmess=fmess,
            agraph=agraph,
            bgraph=bgraph,
            atom_scope=atom_scope,
            bond_scope=bond_scope,
            tgt_token_ids=tgt_token_ids,
            tgt_lengths=tgt_lengths,
            distances=distances
        )

        return g2s_batch

    def __len__(self):
        return len(self.batch_sizes)


class G2SDataset_single(Dataset):
    def __init__(self, feat, vocab):
        self.a_scopes = []
        self.b_scopes = []
        self.a_features = []
        self.b_features = []
        self.a_graphs = []
        self.b_graphs = []
        self.a_scopes_lens = []
        self.b_scopes_lens = []
        self.a_features_lens = []
        self.b_features_lens = []

        self.src_token_ids = []             # loaded but not batched
        self.src_lens = []

        self.data_indices = []
        self.batch_sizes = []
        self.batch_starts = [0]
        self.batch_ends = [1]

        self.vocab = vocab
        self.vocab_tokens = [k for k, v in sorted(self.vocab.items(), key=lambda tup: tup[1])]

        for attr in ["a_scopes", "b_scopes", "a_features", "b_features", "a_graphs", "b_graphs",
                     "a_scopes_lens", "b_scopes_lens", "a_features_lens", "b_features_lens",
                     "src_token_ids", "src_lens"]:
            setattr(self, attr, feat[attr])

        # mask out chiral tag (as UNSPECIFIED)
        self.a_features[:, 6] = 2

        assert len(self.a_scopes_lens) == len(self.b_scopes_lens) == \
               len(self.a_features_lens) == len(self.b_features_lens) == \
               len(self.src_token_ids) == len(self.src_lens), \
               f"Lengths of source and target mismatch!"

        self.a_scopes_indices = len2idx(self.a_scopes_lens)
        self.b_scopes_indices = len2idx(self.b_scopes_lens)
        self.a_features_indices = len2idx(self.a_features_lens)
        self.b_features_indices = len2idx(self.b_features_lens)

        del self.a_scopes_lens, self.b_scopes_lens, self.a_features_lens, self.b_features_lens

        self.data_size = len(self.src_token_ids)
        self.data_indices = np.arange(self.data_size)

    def __getitem__(self, index: int) -> G2SBatch:
        batch_index = index
        batch_start = self.batch_starts[batch_index]
        batch_end = self.batch_ends[batch_index]

        data_indices = self.data_indices[batch_start:batch_end]

        graph_features = []
        a_lengths = []
        for data_index in data_indices:
            start, end = self.a_scopes_indices[data_index]
            a_scope = self.a_scopes[start:end]
            a_length = a_scope[-1][0] + a_scope[-1][1] - a_scope[0][0]

            start, end = self.b_scopes_indices[data_index]
            b_scope = self.b_scopes[start:end]

            start, end = self.a_features_indices[data_index]
            a_feature = self.a_features[start:end]
            a_graph = self.a_graphs[start:end]

            start, end = self.b_features_indices[data_index]
            b_feature = self.b_features[start:end]
            b_graph = self.b_graphs[start:end]

            graph_feature = (a_scope, b_scope, a_feature, b_feature, a_graph, b_graph)
            graph_features.append(graph_feature)
            a_lengths.append(a_length)

        fnode, fmess, agraph, bgraph, atom_scope, bond_scope = collate_graph_features(graph_features)

        distances = collate_graph_distances(graph_features, a_lengths)

        g2s_batch = G2SBatch_single(
            fnode=fnode,
            fmess=fmess,
            agraph=agraph,
            bgraph=bgraph,
            atom_scope=atom_scope,
            bond_scope=bond_scope,
            distances=distances
        )

        return g2s_batch

    def __len__(self):
        return 1


class G2SBatch_single:
    def __init__(self,
                 fnode: torch.Tensor,
                 fmess: torch.Tensor,
                 agraph: torch.Tensor,
                 bgraph: torch.Tensor,
                 atom_scope: List,
                 bond_scope: List,
                 distances: torch.Tensor = None):
        self.fnode = fnode
        self.fmess = fmess
        self.agraph = agraph
        self.bgraph = bgraph
        self.atom_scope = atom_scope
        self.bond_scope = bond_scope
        self.distances = distances
        self.size = 1

    def to(self, device):
        self.fnode = self.fnode.to(device)
        self.fmess = self.fmess.to(device)
        self.agraph = self.agraph.to(device)
        self.bgraph = self.bgraph.to(device)

        if self.distances is not None:
            self.distances = self.distances.to(device)


def get_graph_from_smiles(smi: str):
    mol = Chem.MolFromSmiles(smi)
    rxn_graph = RxnGraph(reac_mol=mol)

    return rxn_graph


def get_graph_features_from_smi(_args):
    i, smi, use_rxn_class = _args
    assert isinstance(smi, str) and isinstance(use_rxn_class, bool)
    if i > 0 and i % 10000 == 0:
        logging.info(f"Processing {i}th SMILES")

    atom_features = []
    bond_features = []
    edge_dict = {}

    if not smi.strip():
        smi = "CC"          # hardcode to ignore

    graph = get_graph_from_smiles(smi).reac_mol

    mol = graph.mol
    assert mol.GetNumAtoms() == len(graph.G_dir)

    G = nx.convert_node_labels_to_integers(graph.G_dir, first_label=0)

    # node iteration to get sparse atom features
    for v, attr in G.nodes(data="label"):
        atom_feat = get_atom_features_sparse(mol.GetAtomWithIdx(v),
                                             use_rxn_class=use_rxn_class,
                                             rxn_class=graph.rxn_class)
        atom_features.append(atom_feat)

    a_graphs = [[] for _ in range(len(atom_features))]

    # edge iteration to get (dense) bond features
    for u, v, attr in G.edges(data='label'):
        bond_feat = get_bond_features(mol.GetBondBetweenAtoms(u, v))
        bond_feat = [u, v] + bond_feat
        bond_features.append(bond_feat)

        eid = len(edge_dict)
        edge_dict[(u, v)] = eid
        a_graphs[v].append(eid)

    b_graphs = [[] for _ in range(len(bond_features))]

    # second edge iteration to get neighboring edges (after edge_dict is updated fully)
    for bond_feat in bond_features:
        u, v = bond_feat[:2]
        eid = edge_dict[(u, v)]

        for w in G.predecessors(u):
            if not w == v:
                b_graphs[eid].append(edge_dict[(w, u)])

    # padding
    for a_graph in a_graphs:
        while len(a_graph) < 11:            # OH MY GOODNESS... Fe can be bonded to 10...
            a_graph.append(1e9)

    for b_graph in b_graphs:
        while len(b_graph) < 11:            # OH MY GOODNESS... Fe can be bonded to 10...
            b_graph.append(1e9)

    a_scopes = np.array(graph.atom_scope, dtype=np.int32)
    a_scopes_lens = a_scopes.shape[0]
    b_scopes = np.array(graph.bond_scope, dtype=np.int32)
    b_scopes_lens = b_scopes.shape[0]
    a_features = np.array(atom_features, dtype=np.int32)
    a_features_lens = a_features.shape[0]
    b_features = np.array(bond_features, dtype=np.int32)
    b_features_lens = b_features.shape[0]
    a_graphs = np.array(a_graphs, dtype=np.int32)
    b_graphs = np.array(b_graphs, dtype=np.int32)

    return a_scopes, a_scopes_lens, b_scopes, b_scopes_lens, \
        a_features, a_features_lens, b_features, b_features_lens, a_graphs, b_graphs


def collate_graph_features(graph_features: List[Tuple], directed: bool = True, use_rxn_class: bool = False) \
        -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, List[np.ndarray], List[np.ndarray]]:
    if directed:
        padded_features = get_atom_features_sparse(Chem.Atom("*"), use_rxn_class=use_rxn_class, rxn_class=0)
        fnode = [np.array(padded_features)]
        fmess = [np.zeros(shape=[1, 2 + BOND_FDIM], dtype=np.int32)]
        agraph = [np.zeros(shape=[1, 11], dtype=np.int32)]
        bgraph = [np.zeros(shape=[1, 11], dtype=np.int32)]

        n_unique_bonds = 1
        edge_offset = 1

        atom_scope, bond_scope = [], []

        for bid, graph_feature in enumerate(graph_features):
            a_scope, b_scope, atom_features, bond_features, a_graph, b_graph = graph_feature

            a_scope = a_scope.copy()
            b_scope = b_scope.copy()
            atom_features = atom_features.copy()
            bond_features = bond_features.copy()
            a_graph = a_graph.copy()
            b_graph = b_graph.copy()

            atom_offset = len(fnode)
            bond_offset = n_unique_bonds
            n_unique_bonds += int(bond_features.shape[0] / 2)              # This should be correct?

            a_scope[:, 0] += atom_offset
            b_scope[:, 0] += bond_offset
            atom_scope.append(a_scope)
            bond_scope.append(b_scope)

            # node iteration is reduced to an extend
            fnode.extend(atom_features)

            # edge iteration is reduced to an append
            bond_features[:, :2] += atom_offset
            fmess.append(bond_features)

            a_graph += edge_offset
            a_graph[a_graph >= 999999999] = 0           # resetting padding edge to point towards edge 0
            agraph.append(a_graph)

            b_graph += edge_offset
            b_graph[b_graph >= 999999999] = 0           # resetting padding edge to point towards edge 0
            bgraph.append(b_graph)

            edge_offset += bond_features.shape[0]

        # densification
        fnode = np.stack(fnode, axis=0)
        fnode_one_hot = np.zeros([fnode.shape[0], sum(ATOM_FDIM)], dtype=np.float32)

        for i in range(len(ATOM_FDIM) - 1):
            fnode[:, i+1:] += ATOM_FDIM[i]          # cumsum, essentially

        for i, feat in enumerate(fnode):            # Looks vectorizable?
            # fnode_one_hot[i, feat[feat < 9999]] = 1
            fnode_one_hot[i, feat[feat < sum(ATOM_FDIM)]] = 1

        fnode = torch.as_tensor(fnode_one_hot, dtype=torch.float)
        fmess = torch.as_tensor(np.concatenate(fmess, axis=0), dtype=torch.float)

        agraph = np.concatenate(agraph, axis=0)
        column_idx = np.argwhere(np.all(agraph[..., :] == 0, axis=0))
        agraph = agraph[:, :column_idx[0, 0] + 1]       # drop trailing columns of 0, leaving only 1 last column of 0

        bgraph = np.concatenate(bgraph, axis=0)
        column_idx = np.argwhere(np.all(bgraph[..., :] == 0, axis=0))
        bgraph = bgraph[:, :column_idx[0, 0] + 1]       # drop trailing columns of 0, leaving only 1 last column of 0

        agraph = torch.as_tensor(agraph, dtype=torch.long)
        bgraph = torch.as_tensor(bgraph, dtype=torch.long)

    else:
        raise NotImplementedError

    return fnode, fmess, agraph, bgraph, atom_scope, bond_scope


def collate_graph_distances(graph_features: List[Tuple], a_lengths: List[int]) -> torch.Tensor:
    max_len = max(a_lengths)

    distances = []
    for bid, (graph_feature, a_length) in enumerate(zip(graph_features, a_lengths)):
        _, _, _, bond_features, _, _ = graph_feature
        bond_features = bond_features.copy()

        # compute adjacency
        adjacency = np.zeros((a_length, a_length), dtype=np.int32)
        for bond_feature in bond_features:
            u, v = bond_feature[:2]
            adjacency[u, v] = 1

        # compute graph distance
        distance = adjacency.copy()
        shortest_paths = adjacency.copy()
        path_length = 2
        stop_counter = 0
        non_zeros = 0

        while 0 in distance:
            shortest_paths = np.matmul(shortest_paths, adjacency)
            shortest_paths = path_length * (shortest_paths > 0)
            new_distance = distance + (distance == 0) * shortest_paths

            # if np.count_nonzero(new_distance) == np.count_nonzero(distance):
            if np.count_nonzero(new_distance) <= non_zeros:
                stop_counter += 1
            else:
                non_zeros = np.count_nonzero(new_distance)
                stop_counter = 0

            distance = new_distance
            path_length += 1

        # bucket
        distance[(distance > 8) & (distance < 15)] = 8
        distance[distance >= 15] = 9

        # reset diagonal
        np.fill_diagonal(distance, 0)

        # padding
        padded_distance = np.ones((max_len, max_len), dtype=np.int32) * 10
        padded_distance[:a_length, :a_length] = distance

        distances.append(padded_distance)

    distances = np.stack(distances)
    distances = torch.as_tensor(distances, dtype=torch.long)

    return distances


def make_vocab(fns: Dict[str, List[Tuple[str, str]]], vocab_file: str, tokenized=True):
    assert tokenized, f"Vocab can only be made from tokenized files"

    logging.info(f"Making vocab from {fns}")
    vocab = {}

    for phase, file_list in fns.items():
        for src_file, tgt_file in file_list:
            for fn in [src_file, tgt_file]:
                with open(fn, "r") as f:
                    for line in f:
                        tokens = line.strip().split()
                        for token in tokens:
                            if token in vocab:
                                vocab[token] += 1
                            else:
                                vocab[token] = 1

    logging.info(f"Saving vocab into {vocab_file}")
    with open(vocab_file, "w") as of:
        of.write("_PAD\n_UNK\n_SOS\n_EOS\n")
        for token, count in vocab.items():
            of.write(f"{token}\t{count}\n")


def load_vocab(vocab_file: str) -> Dict[str, int]:
    vocab = {}
    with open(vocab_file, "r") as f:
        for i, line in enumerate(f):
            token = line.strip().split("\t")[0]
            vocab[token] = i

    return vocab


def data_util_test():
    pass


if __name__ == "__main__":
    data_util_test()



import rdkit.Chem as Chem
import rdkit.Chem.AllChem as AllChem
from rdkit.Chem.rdchem import ChiralType, BondType, BondDir, BondStereo

from rdchiral.chiral import template_atom_could_have_been_tetra
from rdchiral.utils import vprint, PLEVEL
from rdchiral.bonds import enumerate_possible_cistrans_defs, bond_dirs_by_mapnum, \
    get_atoms_across_double_bonds

BondDirOpposite = {AllChem.BondDir.ENDUPRIGHT: AllChem.BondDir.ENDDOWNRIGHT,
                   AllChem.BondDir.ENDDOWNRIGHT: AllChem.BondDir.ENDUPRIGHT}

class rdchiralReaction():
    '''
    Class to store everything that should be pre-computed for a reaction. This
    makes library application much faster, since we can pre-do a lot of work
    instead of doing it for every mol-template pair
    '''
    def __init__(self, reaction_smarts):
        # Keep smarts, useful for reporting
        self.reaction_smarts = reaction_smarts

        # Initialize - assigns stereochemistry and fills in missing rct map numbers
        self.rxn = initialize_rxn_from_smarts(reaction_smarts)

        # Combine template fragments so we can play around with mapnums
        self.template_r, self.template_p = get_template_frags_from_rxn(self.rxn)

        # Define molAtomMapNumber->atom dictionary for template rct and prd
        self.atoms_rt_map = {a.GetAtomMapNum(): a \
            for a in self.template_r.GetAtoms() if a.GetAtomMapNum()}
        self.atoms_pt_map = {a.GetAtomMapNum(): a \
            for a in self.template_p.GetAtoms() if a.GetAtomMapNum()}

        # Back-up the mapping for the reaction
        self.atoms_rt_idx_to_map = {a.GetIdx(): a.GetAtomMapNum()
            for a in self.template_r.GetAtoms()}
        self.atoms_pt_idx_to_map = {a.GetIdx(): a.GetAtomMapNum()
            for a in self.template_p.GetAtoms()}

        # Check consistency (this should not be necessary...)
        if any(self.atoms_rt_map[i].GetAtomicNum() != self.atoms_pt_map[i].GetAtomicNum() \
                for i in self.atoms_rt_map if i in self.atoms_pt_map):
            raise ValueError('Atomic identity should not change in a reaction!')

        # Call template_atom_could_have_been_tetra to pre-assign value to atom
        [template_atom_could_have_been_tetra(a) for a in self.template_r.GetAtoms()]
        [template_atom_could_have_been_tetra(a) for a in self.template_p.GetAtoms()]

        # Pre-list chiral double bonds (for copying back into outcomes/matching)
        self.rt_bond_dirs_by_mapnum = bond_dirs_by_mapnum(self.template_r)
        self.pt_bond_dirs_by_mapnum = bond_dirs_by_mapnum(self.template_p)

        # Enumerate possible cis/trans...
        self.required_rt_bond_defs, self.required_bond_defs_coreatoms = \
            enumerate_possible_cistrans_defs(self.template_r)

    def reset(self):
        for (idx, mapnum) in self.atoms_rt_idx_to_map.items():
            self.template_r.GetAtomWithIdx(idx).SetAtomMapNum(mapnum)
        for (idx, mapnum) in self.atoms_pt_idx_to_map.items():
            self.template_p.GetAtomWithIdx(idx).SetAtomMapNum(mapnum)

class rdchiralReactants():
    '''
    Class to store everything that should be pre-computed for a reactant mol
    so that library application is faster
    '''
    def __init__(self, reactant_smiles):
        # Keep original smiles, useful for reporting
        self.reactant_smiles = reactant_smiles

        # Initialize into RDKit mol
        self.reactants = initialize_reactants_from_smiles(reactant_smiles)

        # Set mapnum->atom dictionary
        # all reactant atoms must be mapped after initialization, so this is safe
        self.atoms_r = {a.GetAtomMapNum(): a for a in self.reactants.GetAtoms()}
        self.idx_to_mapnum = lambda idx: self.reactants.GetAtomWithIdx(idx).GetAtomMapNum()

        # Create copy of molecule without chiral information, used with
        # RDKit's naive runReactants
        self.reactants_achiral = initialize_reactants_from_smiles(reactant_smiles)
        [a.SetChiralTag(ChiralType.CHI_UNSPECIFIED) for a in self.reactants_achiral.GetAtoms()]
        [(b.SetStereo(BondStereo.STEREONONE), b.SetBondDir(BondDir.NONE)) \
            for b in self.reactants_achiral.GetBonds()]

        # Pre-list reactant bonds (for stitching broken products)
        self.bonds_by_mapnum = [
            (b.GetBeginAtom().GetAtomMapNum(), b.GetEndAtom().GetAtomMapNum(), b) \
            for b in self.reactants.GetBonds()
        ]

        # Pre-list chiral double bonds (for copying back into outcomes/matching)
        self.bond_dirs_by_mapnum = {}
        for (i, j, b) in self.bonds_by_mapnum:
            if b.GetBondDir() != BondDir.NONE:
                self.bond_dirs_by_mapnum[(i, j)] = b.GetBondDir()
                self.bond_dirs_by_mapnum[(j, i)] = BondDirOpposite[b.GetBondDir()]

        # Get atoms across double bonds defined by mapnum
        self.atoms_across_double_bonds = get_atoms_across_double_bonds(self.reactants)


def initialize_rxn_from_smarts(reaction_smarts):
    # Initialize reaction
    rxn = AllChem.ReactionFromSmarts(reaction_smarts)
    rxn.Initialize()
    if rxn.Validate()[1] != 0:
        raise ValueError('validation failed')
    if PLEVEL >= 2: print('Validated rxn without errors')


    # Figure out if there are unnecessary atom map numbers (that are not balanced)
    # e.g., leaving groups for retrosynthetic templates. This is because additional
    # atom map numbers in the input SMARTS template may conflict with the atom map
    # numbers of the molecules themselves
    prd_maps = [a.GetAtomMapNum() for prd in rxn.GetProducts() for a in prd.GetAtoms() if a.GetAtomMapNum()]

    unmapped = 700
    for rct in rxn.GetReactants():
        rct.UpdatePropertyCache(strict=False)
        Chem.AssignStereochemistry(rct)
        # Fill in atom map numbers
        for a in rct.GetAtoms():
            if not a.GetAtomMapNum() or a.GetAtomMapNum() not in prd_maps:
                a.SetAtomMapNum(unmapped)
                unmapped += 1
    if PLEVEL >= 2: print('Added {} map nums to unmapped reactants'.format(unmapped-700))
    if unmapped > 800:
        raise ValueError('Why do you have so many unmapped atoms in the template reactants?')

    return rxn

def initialize_reactants_from_smiles(reactant_smiles):
    # Initialize reactants
    reactants = Chem.MolFromSmiles(reactant_smiles)
    Chem.AssignStereochemistry(reactants, flagPossibleStereoCenters=True)
    reactants.UpdatePropertyCache(strict=False)
    # To have the product atoms match reactant atoms, we
    # need to populate the map number field, since this field
    # gets copied over during the reaction via reactant_atom_idx.
    [a.SetAtomMapNum(i+1) for (i, a) in enumerate(reactants.GetAtoms())]
    if PLEVEL >= 2: print('Initialized reactants, assigned map numbers, stereochem, flagpossiblestereocenters')
    return reactants

def get_template_frags_from_rxn(rxn):
    # Copy reaction template so we can play around with map numbers
    for i, rct in enumerate(rxn.GetReactants()):
        if i == 0:
            template_r = rct
        else:
            template_r = AllChem.CombineMols(template_r, rct)
    for i, prd in enumerate(rxn.GetProducts()):
        if i == 0:
            template_p = prd
        else:
            template_p = AllChem.CombineMols(template_p, prd)
    return template_r, template_p


from rdkit import Chem
from typing import List


# Symbols for different atoms
ATOM_LIST = ['C', 'N', 'O', 'S', 'F', 'Si', 'P', 'Cl', 'Br', 'Mg', 'Na', 'Ca', 'Fe',
             'As', 'Al', 'I', 'B', 'V', 'K', 'Tl', 'Yb', 'Sb', 'Sn', 'Ag', 'Pd', 'Co', 'Se', 'Ti',
             'Zn', 'H', 'Li', 'Ge', 'Cu', 'Au', 'Ni', 'Cd', 'In', 'Mn', 'Zr', 'Cr', 'Pt', 'Hg', 'Pb',
             'W', 'Ru', 'Nb', 'Re', 'Te', 'Rh', 'Ta', 'Tc', 'Ba', 'Bi', 'Hf', 'Mo', 'U', 'Sm', 'Os', 'Ir',
             'Ce', 'Gd', 'Ga', 'Cs', '*', 'unk']
ATOM_DICT = {symbol: i for i, symbol in enumerate(ATOM_LIST)}

MAX_NB = 10
DEGREES = list(range(MAX_NB))
HYBRIDIZATION = [Chem.rdchem.HybridizationType.SP,
                 Chem.rdchem.HybridizationType.SP2,
                 Chem.rdchem.HybridizationType.SP3,
                 Chem.rdchem.HybridizationType.SP3D,
                 Chem.rdchem.HybridizationType.SP3D2]
HYBRIDIZATION_DICT = {hb: i for i, hb in enumerate(HYBRIDIZATION)}

FORMAL_CHARGE = [-1, -2, 1, 2, 0]
FC_DICT = {fc: i for i, fc in enumerate(FORMAL_CHARGE)}

VALENCE = [0, 1, 2, 3, 4, 5, 6]
VALENCE_DICT = {vl: i for i, vl in enumerate(VALENCE)}

NUM_Hs = [0, 1, 3, 4, 5]
NUM_Hs_DICT = {nH: i for i, nH in enumerate(NUM_Hs)}

CHIRAL_TAG = [Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
              Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
              Chem.rdchem.ChiralType.CHI_UNSPECIFIED]
CHIRAL_TAG_DICT = {ct: i for i, ct in enumerate(CHIRAL_TAG)}

RS_TAG = ["R", "S", "None"]
RS_TAG_DICT = {rs: i for i, rs in enumerate(RS_TAG)}

BOND_TYPES = [None,
              Chem.rdchem.BondType.SINGLE,
              Chem.rdchem.BondType.DOUBLE,
              Chem.rdchem.BondType.TRIPLE,
              Chem.rdchem.BondType.AROMATIC]
BOND_FLOAT_TO_TYPE = {
    0.0: BOND_TYPES[0],
    1.0: BOND_TYPES[1],
    2.0: BOND_TYPES[2],
    3.0: BOND_TYPES[3],
    1.5: BOND_TYPES[4],
}

BOND_STEREO = [Chem.rdchem.BondStereo.STEREOE,
               Chem.rdchem.BondStereo.STEREOZ,
               Chem.rdchem.BondStereo.STEREONONE]

BOND_DELTAS = {-3: 0, -2: 1, -1.5: 2, -1: 3, -0.5: 4, 0: 5, 0.5: 6, 1: 7, 1.5: 8, 2: 9, 3: 10}
BOND_FLOATS = [0.0, 1.0, 2.0, 3.0, 1.5]

RXN_CLASSES = list(range(10))

# ATOM_FDIM = len(ATOM_LIST) + len(DEGREES) + len(FORMAL_CHARGE) + len(HYBRIDIZATION) \
#             + len(VALENCE) + len(NUM_Hs) + 1
ATOM_FDIM = [len(ATOM_LIST), len(DEGREES), len(FORMAL_CHARGE), len(HYBRIDIZATION), len(VALENCE),
             len(NUM_Hs), len(CHIRAL_TAG), len(RS_TAG), 2]
# BOND_FDIM = 6
BOND_FDIM = 9
BINARY_FDIM = 5 + BOND_FDIM
INVALID_BOND = -1


def get_atom_features_sparse(atom: Chem.Atom, rxn_class: int = None, use_rxn_class: bool = False) -> List[int]:
    """Get atom features as sparse idx.

    Parameters
    ----------
    atom: Chem.Atom,
        Atom object from RDKit
    rxn_class: int, None
        Reaction class the molecule was part of
    use_rxn_class: bool, default False,
        Whether to use reaction class as additional input
    """
    feature_array = []
    symbol = atom.GetSymbol()
    symbol_id = ATOM_DICT.get(symbol, ATOM_DICT["unk"])
    feature_array.append(symbol_id)

    if symbol in ["*", "unk"]:
        padding = [999999999] * len(ATOM_FDIM) if use_rxn_class else [999999999] * (len(ATOM_FDIM) - 1)
        feature_array.extend(padding)

    else:
        degree_id = atom.GetDegree()
        if degree_id not in DEGREES:
            degree_id = 9
        formal_charge_id = FC_DICT.get(atom.GetFormalCharge(), 4)
        hybridization_id = HYBRIDIZATION_DICT.get(atom.GetHybridization(), 4)
        valence_id = VALENCE_DICT.get(atom.GetTotalValence(), 6)
        num_h_id = NUM_Hs_DICT.get(atom.GetTotalNumHs(), 4)
        chiral_tag_id = CHIRAL_TAG_DICT.get(atom.GetChiralTag(), 2)

        rs_tag = atom.GetPropsAsDict().get("_CIPCode", "None")
        rs_tag_id = RS_TAG_DICT.get(rs_tag, 2)

        is_aromatic = int(atom.GetIsAromatic())
        feature_array.extend([degree_id, formal_charge_id, hybridization_id,
                              valence_id, num_h_id, chiral_tag_id, rs_tag_id, is_aromatic])

        if use_rxn_class:
            feature_array.append(rxn_class)

    return feature_array


def get_bond_features(bond: Chem.Bond) -> List[int]:
    """Get bond features.

    Parameters
    ----------
    bond: Chem.Bond,
        bond object
    """
    bt = bond.GetBondType()
    bond_features = [int(bt == bond_type) for bond_type in BOND_TYPES[1:]]
    bs = bond.GetStereo()
    bond_features.extend([int(bs == bond_stereo) for bond_stereo in BOND_STEREO])
    bond_features.extend([int(bond.GetIsConjugated()), int(bond.IsInRing())])

    return bond_features



from setuptools import setup, find_packages
# import setuptools

"torch@https://download.pytorch.org/whl/torch/torch-1.12.0+cu113-cp310-cp310-linux_x86_64.whl"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="readretro_test",
    version="1.0.10",
    author="Taein Kim",
    author_email="tedori725@kaist.ac.kr",
    description="READRetro lib test",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package=find_packages(),
    install_requires=[
        'h5py',
        'typing-extensions',
        'wheel',
        'easydict',
        'pandas',
        'tqdm',
        'numpy==1.22',
        'OpenNMT-py==2.3.0',
        'networkx==2.5',
        'torch==1.12.0'
    ],
    url="https://github.com/SeulLee05/READRetro",
    project_urls={
        "Bug Tracker": "https://github.com/SeulLee05/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    # packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
)